# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the data interfaces."""

import socket
import sys
import unittest
from copy import deepcopy

import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.data import (DownloadedMarketData, Fred,
                               UserProvidedMarketData, YahooFinance,
                               _loader_csv, _loader_pickle, _loader_sqlite,
                               _storer_csv, _storer_pickle, _storer_sqlite)
from cvxportfolio.errors import DataError
from cvxportfolio.tests import CvxportfolioTest


class NoInternet():
    """Context with no internet.

    Adapted from: https://github.com/dvl/python-internet-sabotage
    """
    _module = sys.modules[__name__]

    def __init__(self, exception=IOError):
        self.exception = exception

    def _enable_socket(self):
        """Enable sockets in this module."""
        setattr(self._module, '_socket_disabled', False)

    def _disable_socket(self):
        """Disable sockets in this module."""
        setattr(self._module, '_socket_disabled', True)

        def _guarded(*args, **kwargs):
            """Monkey-patch the socket module."""
            if getattr(self._module, '_socket_disabled', False):
                raise self.exception('Internet is disabled')

            return socket.SocketType(*args, **kwargs)

        socket.socket = _guarded

    def __enter__(self):
        """Open context."""
        self._disable_socket()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close context, discard exceptions."""
        self._enable_socket()


class TestData(CvxportfolioTest):
    """Test SymbolData methods and interface."""

    def test_yfinance_download(self):
        """Test YfinanceBase."""

        storer = YahooFinance('AAPL', base_location=self.datadir)

        # pylint: disable=protected-access
        data = storer._download("AAPL", start="2023-04-01", end="2023-04-15")

        self.assertTrue(np.isclose(
            data.loc["2023-04-10 13:30:00+00:00", "return"],
            data.loc["2023-04-11 13:30:00+00:00", "open"] /
            data.loc["2023-04-10 13:30:00+00:00", "open"] - 1,
            rtol=1e-04, atol=1e-07,
        ))

    def test_fred(self):
        """Test basic Fred usage."""

        store = Fred(
            symbol="DFF", storage_backend='pickle',
            base_location=self.datadir)

        print(store.data)
        data = store.data
        self.assertTrue(np.isclose(data["2023-04-10"], 4.83))
        self.assertTrue(data.index[0] ==
            pd.Timestamp("1954-07-01 00:00:00+00:00"))

        # test update
        olddata = pd.Series(data.iloc[:-123], copy=True)
        olddata.index = olddata.index.tz_localize(None)
        # pylint: disable=protected-access
        newdata = store._preload(store._download("DFF", olddata))
        self.assertTrue(np.all(store.data == newdata))

        # test not re-downloading
        _ = Fred(
            symbol="DFF", grace_period=pd.Timedelta('10d'),
            base_location=self.datadir)

    def test_yahoo_finance(self):
        """Test yahoo finance ability to store and retrieve."""

        store = YahooFinance(
            symbol="ZM", storage_backend='pickle',
            base_location=self.datadir)

        data = store.data

        # print(data)

        self.assertTrue(np.isclose(
            data.loc["2023-04-05 13:30:00+00:00", "return"],
            data.loc["2023-04-06 13:30:00+00:00", "open"] /
            data.loc["2023-04-05 13:30:00+00:00", "open"] - 1,
            rtol=1e-04, atol=1e-07,
        ))

        store.update(grace_period=pd.Timedelta('1d'))
        data1 = store.load()
        # print(data1)

        self.assertTrue(np.isnan(data1.iloc[-1]["close"]))

        # print((data1.iloc[: len(data) - 1].Return -
        #       data.iloc[:-1].Return).describe().T)

        self.assertTrue(np.allclose(
            data1.loc[data.index[:-1]]['return'], data.iloc[:-1]['return']))

        # test not re-downloading
        _ = YahooFinance(
            symbol="ZM", grace_period=pd.Timedelta('10d'),
            base_location=self.datadir)

    def test_yahoo_finance_removefirstline(self):
        """Test that the first line of OHLCV is removed if there are NaNs."""

        # this symbol was found to have NaNs in the first line
        _ = YahooFinance(
            symbol="CVX", storage_backend='pickle',
            base_location=self.datadir)

    @unittest.skipIf(sys.version_info.major == 3
        and sys.version_info.minor < 11, "Issues with timezoned timestamps.")
    def test_sqlite3_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        with self.assertWarns(UserWarning):
            self._base_test_series(_loader_sqlite, _storer_sqlite)

    @unittest.skipIf(sys.version_info.major == 3
        and sys.version_info.minor < 11, "Issues with timezoned timestamps.")
    def test_local_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self._base_test_series(_loader_csv, _storer_csv)

    def test_pickle_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self._base_test_series(_loader_pickle, _storer_pickle)

    def test_sqlite3_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self._base_test_dataframe(_loader_sqlite, _storer_sqlite)

    def test_local_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self._base_test_dataframe(_loader_csv, _storer_csv)

    def test_pickle_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self._base_test_dataframe(_loader_pickle, _storer_pickle)

    def test_local_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self._base_test_multiindex(_loader_csv, _storer_csv)

    def test_sqlite3_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self._base_test_multiindex(_loader_sqlite, _storer_sqlite)

    def test_pickle_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self._base_test_multiindex(_loader_pickle, _storer_pickle)

    def _base_test_series(self, loader, storer):
        """Test storing and retrieving of a Series with datetime index."""

        for data in [
            pd.Series(
                0.0, pd.date_range("2020-01-01", "2020-01-10"),
                name="test0"),
            pd.Series(
                0.0, pd.date_range("2020-01-01", "2020-01-10", tz='UTC-05:00'),
                name="test1"),
            pd.Series(
                3, pd.date_range("2020-01-01", "2020-01-10", tz='UTC'),
                name="test2"),
            pd.Series("hello",
                pd.date_range("2020-01-01", "2020-01-02",  tz='UTC-05:00',
                    freq="h"),
                name="test3"),
            # test overwrite
            pd.Series("hello",
                pd.date_range("2020-01-01", "2020-01-02",  tz='UTC', freq="h"),
                name="test3"),
            # test datetime conversion
            pd.Series(
                pd.date_range("2022-01-01", "2022-01-02",  tz='UTC',
                    freq="h"),
                pd.date_range("2020-01-01", "2020-01-02",  tz='UTC', freq="h"),
                name="test4"),
            ]:

            # print(data)
            # print(data.index)
            # print(data.index[0])
            # print(data.index[0].tzinfo)
            # print(data.index.dtype)
            # print(data.dtypes)

            storer(data.name, data, self.datadir)

            data1 = loader(data.name, self.datadir)
            # print(data1)
            # print(data1.index)
            # print(data1.index[0])
            # print(data1.index[0].tzinfo)
            # print(data1.index.dtype)
            # print(data1.dtypes)

            self.assertTrue(data.name == data1.name)
            self.assertTrue(all(data == data1))
            self.assertTrue(all(data.index == data1.index))
            self.assertTrue(data.dtypes == data1.dtypes)

        # test load not existent
        try:
            self.assertTrue(loader('blahblah', self.datadir) is None)
        except FileNotFoundError:
            pass

    def _base_test_dataframe(self, loader, storer):
        """Test storing and retrieving of a DataFrame with datetime index."""

        index = pd.date_range("2020-01-01", "2020-01-02", freq="h", tz='UTC')
        data = {
            "one": range(len(index)),
            "two": np.arange(len(index)) / 19.0,
            "three": ["hello"] * len(index),
            "four": [np.nan] * len(index),
        }

        data["two"][2] = np.nan
        data = pd.DataFrame(data, index=index)
        # print(data)
        # print(data.index.dtype)
        # print(data.dtypes)

        storer("example", data, self.datadir)
        data1 = loader("example", self.datadir)
        # print(data1)
        # print(data1.index.dtype)
        # print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.dtypes == data1.dtypes))

    def _base_test_multiindex(self, loader, storer):
        """Test storing and retrieving of a Series or DataFrame with multi-.

        index.
        """
        # second level is object
        timeindex = pd.date_range("2022-01-01", "2022-01-30", tz='UTC')
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 3), index=index)
        data.columns = ["one", "two", "tre"]

        # print(data.index)
        # print(data)
        # print(data.index.dtype)
        # print(data.dtypes)

        storer("example", data, self.datadir)
        data1 = loader("example", self.datadir)

        # print(data1.index)
        # print(data1)
        # print(data1.index.dtype)
        # print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.index.dtypes == data1.index.dtypes))
        self.assertTrue(all(data.dtypes == data1.dtypes))

        # second level is timestamp
        timeindex = pd.date_range("2022-01-01", "2022-01-30", tz='UTC')
        second_level = pd.date_range("2022-01-01", "2022-01-03", tz='UTC')
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 3), index=index)
        data.columns = ["a", "b", "c"]

        #print(data.index)
        # print(data)
        # print(data.index.dtypes)
        # print(data.dtypes)

        storer("example", data, self.datadir)
        data1 = loader("example", self.datadir)

        #print(data1.index)
        # print(data1)
        # print(data1.index.dtypes)
        # print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.index.dtypes == data1.index.dtypes))
        self.assertTrue(all(data.dtypes == data1.dtypes))

    def test_download_errors(self):
        """Test single-symbol download error."""

        storer = YahooFinance(
            'AAPL', grace_period=self.data_grace_period,
            base_location=self.datadir)
        # with self.assertRaises(SyntaxError):
        #     # pylint: disable=protected-access
        #     storer._download('AAPL', overlap=1)

        class YahooFinanceErroneous(YahooFinance):
            """Modified YF that nans last open price."""
            def _download(
                    self, symbol, current=None, grace_period='5d', **kwargs):
                """Modified download method."""
                res = super()._download(symbol, current,
                    grace_period=grace_period)
                res.iloc[-1, 0 ] = np.nan
                return res

        _ = YahooFinanceErroneous('AMZN', base_location=self.datadir)
        with self.assertLogs(level='WARNING') as _:
            YahooFinanceErroneous(
                'AMZN', base_location=self.datadir)
            self.assertTrue(np.any([
                'changed last open price' in el for el in _.output]))

        class YahooFinanceErroneous2(YahooFinance):
            """Modified YF that nans some line."""
            def _download(
                    self, symbol, current=None, grace_period='5d', **kwargs):
                """Modified download method."""
                res = super()._download(symbol, current,
                    grace_period=grace_period)
                res.iloc[-20] = np.nan
                return res
        with self.assertLogs(level='WARNING') as _:
            _ = YahooFinanceErroneous2('GOOGL',
                base_location=self.datadir)
        with self.assertLogs(level='WARNING') as _:
            _ = YahooFinanceErroneous2(
                'GOOGL', base_location=self.datadir)

        class FredErroneous(Fred):
            """Modified FRED SymbolData that gives a NaN in the last entry."""

            def _download(self, symbol, current, grace_period):
                """Modified download method."""
                res = super()._download(symbol, current,
                    grace_period=grace_period)
                res.iloc[-1] = np.nan
                return res

        _ = FredErroneous('DFF', base_location=self.datadir)
        with self.assertLogs(level='WARNING') as _:
            FredErroneous(
                'DFF', base_location=self.datadir)
            self.assertTrue(np.any([
                'changed last value' in el for el in _.output]))

        class YahooFinanceErroneous3(YahooFinance):
            """Modified YF that is not append-only."""
            counter = 0
            def _download(
                    self, symbol, current=None, grace_period='5d', **kwargs):
                """Modified download method."""
                res = super()._download(symbol, current,
                    grace_period=grace_period)
                if self.counter > 0:
                    res.iloc[-2] = 0.
                self.counter += 1
                return res
        storer = YahooFinanceErroneous3('GOOGL', base_location=self.datadir)
        with self.assertLogs(level='WARNING') as _:
            storer.update(pd.Timedelta('0d'))

    def test_no_internet(self):
        """Test errors thrown when not connected to the internet."""

        with NoInternet():
            with self.assertRaises(DataError):
                cvx.YahooFinance('BABA', base_location=self.datadir)

        with NoInternet():
            with self.assertRaises(DataError):
                cvx.Fred('CES0500000003', base_location=self.datadir)

    def test_yahoo_finance_errors(self):
        """Test simple errors with Yahoo Finance."""

        with self.assertRaises(DataError):
            YahooFinance("DOESNTEXIST", base_location=self.datadir)

    def test_yahoo_finance_update(self):
        """Test specific issues when updating already stored data."""

        raw_data = pd.DataFrame(
            # skip last day because there might actually be issues
            # that invalidate the tests assumptions
            cvx.YahooFinance._get_data_yahoo('AAPL'), copy=True).iloc[:-1]

        class YahooFinanceUpdaterTest(cvx.YahooFinance):
            """Tester of issues with update."""

            def _get_data_yahoo(
                    self, symbol, start=None, *args, **kwargs):
                return pd.DataFrame(
                    self.mock_data if start is None else
                    self.mock_data.loc[self.mock_data.index >= start], copy=True)

            @classmethod
            def _set_mock_data(cls, mock_data):
                cls.mock_data = mock_data

            def _delete_recent_stored(self, how_many):
                raw_stored = self._load_raw()
                only_past = pd.DataFrame(
                    raw_stored.iloc[:-how_many], copy=True)
                only_past.iloc[-1, -1] = np.nan
                self._store(only_past)

        YahooFinanceUpdaterTest._set_mock_data(raw_data)

        obj = YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # warm-up
        initial = YahooFinanceUpdaterTest(
            'AAPL', base_location=self.datadir).data
        updated = YahooFinanceUpdaterTest(
            'AAPL', base_location=self.datadir,
            grace_period=pd.Timedelta('0d')).data
        self.assertTrue(np.allclose(initial, updated, equal_nan=True))
        obj._delete_recent_stored(10)
        re_updated = YahooFinanceUpdaterTest(
            'AAPL', base_location=self.datadir,
            grace_period=pd.Timedelta('0d')).data
        self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))
        print(initial)

        # change intraday data keeping it valid, nothing happens
        raw_data_intraday_changed = pd.DataFrame(raw_data, copy=True)
        raw_data_intraday_changed.iloc[-1, 1] *= .9 # low
        raw_data_intraday_changed.iloc[-1, 2] *= 1.1 # high
        raw_data_intraday_changed.iloc[-1, 3] *= 1.05 # close
        raw_data_intraday_changed.iloc[-1, 4] *= 1.05 # adjclose
        YahooFinanceUpdaterTest._set_mock_data(raw_data_intraday_changed)
        re_updated = YahooFinanceUpdaterTest(
            'AAPL', base_location=self.datadir,
            grace_period=pd.Timedelta('0d')).data
        self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))

        # invalidate last open, gets filled w/ last close
        raw_data_open_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_open_invalid.iloc[-1, 0] = np.inf
        YahooFinanceUpdaterTest._set_mock_data(raw_data_open_invalid)
        with self.assertLogs(level='WARNING') as _:
            re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
            self.assertTrue(np.any(
                ['changed last open price' in el for el in _.output]))
        # print(re_updated)
        self.assertTrue(re_updated.iloc[-1, 0] == re_updated.iloc[-2, 3])

        # reset to init
        obj._delete_recent_stored(20)
        YahooFinanceUpdaterTest._set_mock_data(raw_data)
        YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # make 3 days ago open invalid, nothing changes
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-3, 0] *= 100
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
        self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))

        # reset to init
        obj._delete_recent_stored(20)
        YahooFinanceUpdaterTest._set_mock_data(raw_data)
        YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # make 3 days ago high invalid, nothing changes
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-3, 2] /= 100
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
        self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))

        # make 2 days ago open invalid, stuff changes
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-2, 0] *= 100
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        with self.assertLogs(level='WARNING') as _:
            re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
            self.assertTrue(np.any(
                ['not append-only' in el for el in _.output]))
        # print(re_updated)
        self.assertTrue(re_updated.iloc[-2, 0] == re_updated.iloc[-3, 3])
        # self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))

        # reset to init
        obj._delete_recent_stored(20)
        YahooFinanceUpdaterTest._set_mock_data(raw_data)
        YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # make 3 days ago high invalid, nothing changes
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-3, 2] *= 1000
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
        self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))

        # reset to init
        obj._delete_recent_stored(20)
        YahooFinanceUpdaterTest._set_mock_data(raw_data)
        YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # make 2 days ago close invalid, stuff changes
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-2, 3] *= 1000
        # important: also adjclose needs to wrong in the same way,
        # this is generally true with Yahoo finance
        raw_data_recent_invalid.iloc[-2, 4] *= 1000
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        with self.assertLogs(level='INFO') as _:
            re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
            self.assertTrue(np.any(
                ['not append-only' in el for el in _.output]))
            # for el in _.output:
            #     print(el)
        # print(re_updated)
        self.assertTrue(re_updated.iloc[-2, 0] == re_updated.iloc[-1, 0])
        self.assertFalse(np.allclose(initial, re_updated, equal_nan=True))

        # reset to init
        obj._delete_recent_stored(20)
        YahooFinanceUpdaterTest._set_mock_data(raw_data)
        YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # worse still, invalidate first adjclose of recent data, no ffill
        # data will be eliminated and only last 3 rows will available
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-5, 4] *= 100
        raw_data_recent_invalid.iloc[-5, 3] *= 100
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        with self.assertLogs(level='INFO') as _:
            re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
            # for el in _.output:
            #     print(el)
            self.assertTrue(np.any(
                ['is eliminating data' in el for el in _.output]))

        # print(re_updated)
        self.assertTrue(np.allclose(initial, re_updated, equal_nan=True))

        # reset to init
        obj._delete_recent_stored(20)
        YahooFinanceUpdaterTest._set_mock_data(raw_data)
        YahooFinanceUpdaterTest('AAPL', base_location=self.datadir)

        # even worse, invalidate adjclose 3 and 4 days ago, no ffill
        # data will be eliminated and won't be able to concatenate
        # also change close otherwise dividends check kicks in
        raw_data_recent_invalid = pd.DataFrame(raw_data, copy=True)
        raw_data_recent_invalid.iloc[-3, 4] *= 100
        raw_data_recent_invalid.iloc[-4, 4] *= 100
        raw_data_recent_invalid.iloc[-3, 3] *= 100
        raw_data_recent_invalid.iloc[-4, 3] *= 100
        YahooFinanceUpdaterTest._set_mock_data(
            raw_data_recent_invalid)
        with self.assertLogs(level='INFO') as _:
            re_updated = YahooFinanceUpdaterTest(
                'AAPL', base_location=self.datadir,
                grace_period=pd.Timedelta('0d')).data
            self.assertTrue(np.any(
                ['is eliminating data' in el for el in _.output]))
            self.assertTrue(np.any(
                ['re-downloading from the start' in el for el in _.output]))
            # for el in _.output:
            #     print(el)
        # print(re_updated)
        # last open hasn't changed (re-download is with same bad data, but
        # ffill will work on fresh re-download)
        self.assertTrue(initial.iloc[-1, 0] == re_updated.iloc[-1, 0])

    def test_yahoo_finance_cleaning(self):
        """Test our logic to clean Yahoo Finance data."""

        # this stock was found to have NaN issues
        data = YahooFinance("ENI.MI", base_location=self.datadir).data
        self.assertTrue((data.valuevolume == 0).sum() > 0)
        self.assertTrue(data.iloc[:-1].isnull().sum().sum() == 0)

        # this stock was found to have phony open/low/high prices
        data = YahooFinance('NWG.L', base_location=self.datadir).data
        self.assertGreater(data['return'].min(), -0.75)
        self.assertLess(data['return'].max(), 0.75)

        # this stock had some extreme returns but they were legitimate
        data = YahooFinance('GME', base_location=self.datadir).data
        self.assertGreater(data['return'].min(), -0.75)
        self.assertGreater(data['return'].max(), 3)

    def test_yahoo_finance_preload_warnings(self):
        """Test warnings on _preload if data has issues."""

        # pylint: disable=protected-access

        raw_data = YahooFinance._get_data_yahoo('ZM')
        empty_instance = YahooFinance.__new__(YahooFinance)
        empty_instance._symbol = 'ZM' # because the warnings use the symbol
        cleaned = empty_instance._process(raw_data, None)

        def _test_warning(
                data_transformation, part_of_message, level='WARNING'):
            """Test that warning is raised w/ message containing some word."""
            data = pd.DataFrame(cleaned, copy=True)
            exec(data_transformation) # pylint: disable=exec-used
            # print(data)
            with self.assertLogs(level=level) as _:
                empty_instance._preload(data)
                # print(_)
                self.assertTrue(part_of_message in _.output[0])

        # columns are: open low high close volume return

        # high unexpected return
        _test_warning(
            'data.iloc[300,-1] = 4',
            'dubious total open-to-open returns')

        # low unexpected return
        _test_warning(
            'data.iloc[300,-1] = -0.9',
            'dubious total open-to-open returns')

        # low unexpected open
        _test_warning(
            'data.iloc[300,0] = data.iloc[300,0]*0.1',
            'dubious open to close returns',
            level='INFO')

        # high unexpected open
        _test_warning(
            'data.iloc[300,0] = data.iloc[300,0]*5',
            'dubious open to close returns',
            level='INFO')

        # low unexpected low
        _test_warning(
            'data.iloc[300,1] = data.iloc[300,1]*0.1',
            'dubious open to low returns',
            level='INFO')

        # high unexpected high
        _test_warning(
            'data.iloc[300,2] = data.iloc[300,2]*5',
            'dubious open to high returns',
            level='INFO')

    def test_yahoo_finance_remove_on_many_bad_adjcloses(self):
        """Test remove old data when many adjcloses are invalid."""

        # this stock was found to have bad (negative) adjcloses for many
        # months at its start
        with self.assertLogs(level='INFO') as _:
            YahooFinance('BATS.L', base_location=self.datadir)
            self.assertTrue(np.any(
                'contiguous' in el.output for el in _))

    def test_adjcloses_logrets_removal(self):
        """Test method to remove adjcloses when its logrets are anomalous."""

        # this stock had anomalous price changes in the 70s
        with self.assertLogs(level='INFO') as _:
            d = YahooFinance("SMT.L", base_location=self.datadir).data
            self.assertTrue(np.any([
                    'anomalous adjclose prices' in el for el in _.output]))
            self.assertTrue(d['return'].max() < 2)

        # this stock was found to have phony adjcloses
        with self.assertLogs(level='INFO') as _:
            YahooFinance('BA.L', base_location=self.datadir)
            self.assertTrue(np.any([
                    'anomalous adjclose prices' in el for el in _.output]))

        with self.assertLogs(level='INFO') as _:
            YahooFinance('BA.L', base_location=self.datadir)
            self.assertFalse(np.any([
                    'anomalous adjclose prices' in el for el in _.output]))

    def test_yahoo_finance_cleaning_granular(self):
        """Test each step of cleaning."""

        # pylint: disable=protected-access
        raw_data = YahooFinance._get_data_yahoo('ZM')
        # print(raw_data)
        empty_instance = YahooFinance.__new__(YahooFinance)
        empty_instance._symbol = 'ZM' # because the warnings use the symbol

        def _test_warning(
                data_transformation, part_of_message, level='WARNING'):
            """Test that warning is raised w/ message containing some word."""
            data = pd.DataFrame(raw_data, copy=True)
            exec(data_transformation) # pylint: disable=exec-used
            with self.assertLogs(level=level) as _:
                _cleaned = empty_instance._process(data, None)
                self.assertTrue(
                    np.any([part_of_message in el for el in _.output]))
                # check all NaNs have been filled
                self.assertTrue(_cleaned.iloc[:-1].isnull().sum().sum() == 0)

        def _test_warning_update(
                data_transformation, part_of_message, level='WARNING'):
            """Test that warning is raised w/ message containing some word."""
            new_data = pd.DataFrame(raw_data.iloc[-20:], copy=True)
            saved_data = pd.DataFrame(raw_data.iloc[:-15], copy=True)
            # no need to make it precise
            saved_data['return'] = np.log(saved_data.adjclose).diff()
            del saved_data['adjclose']
            exec(data_transformation) # pylint: disable=exec-used
            with self.assertLogs(level=level) as _:
                _cleaned = empty_instance._process(new_data, saved_data)
                self.assertTrue(
                    np.any([part_of_message in el for el in _.output]))
                # check all NaNs have been filled
                self.assertTrue(_cleaned.iloc[:-1].isnull().sum().sum() == 0)

        # missing initial adjclose
        _test_warning(
            'data.iloc[0,-2] = np.nan',
            'adjclose price is missing', level='INFO')

        # infty
        _test_warning(
            'data.iloc[2,2] = np.inf',
            'infinity', level='INFO')

        # non-pos price
        _test_warning(
            'data.iloc[2,0] = -1',
            'non-positive open', level='INFO')
        _test_warning(
            'data.iloc[2,0] = 0',
            'non-positive open', level='INFO')
        _test_warning(
            'data.iloc[4,2] = 0',
            'non-positive high', level='INFO')

        # neg volume
        _test_warning(
            'data.iloc[2,-1] = -1',
            'negative volumes', level='INFO')

        # open lower low
        _test_warning(
            'data.iloc[1,0] = data.iloc[1,1]*.9',
            'open price lower than low price', level='INFO')

        # open higher high
        _test_warning(
            'data.iloc[1,0] = data.iloc[1,2]*1.1',
            'open price higher than high price', level='INFO')

        # low higher close
        _test_warning(
            'data.iloc[3,1] = data.iloc[3].close * 1.1',
            'low price higher than close price', level='INFO')

        # high lower close
        _test_warning( # had to fix it otherwise open cleaner kicks in
            'close = data.iloc[3].close;'
            'data.iloc[3,0] = close * .95;' # open
            'data.iloc[3,1] = close * .95;' # low
            'data.iloc[3,2] = close * .975', # high
            'high price lower than close price', level='INFO')

        # extreme low price
        _test_warning(
            'data.iloc[3,1] = data.iloc[3,1] * .01',
            'anomalous low price', level='INFO')
        _test_warning(
            'data.iloc[3,1] = data.iloc[3,1] * .02',
            'anomalous low price', level='INFO')
        _test_warning(
            'data.iloc[3,1] = data.iloc[3,1] * .05',
            'anomalous low price', level='INFO')
        _test_warning(
            'data.iloc[3,1] = data.iloc[3,1] * .1',
            'anomalous low price', level='INFO')
        _test_warning(
            'data.iloc[3,1] = data.iloc[3,1] * .2',
            'anomalous low price', level='INFO')
        _test_warning( # changed dtindex until found one that works
            'data.iloc[20,1] = data.iloc[20,1] * .5',
            'anomalous low price', level='INFO')

        # extreme high price
        _test_warning(
            'data.iloc[3,2] = data.iloc[3,2] * 100',
            'anomalous high price', level='INFO')
        _test_warning(
            'data.iloc[3,2] = data.iloc[3,2] * 50',
            'anomalous high price', level='INFO')
        _test_warning(
            'data.iloc[3,2] = data.iloc[3,2] * 20',
            'anomalous high price', level='INFO')
        _test_warning(
            'data.iloc[3,2] = data.iloc[3,2] * 10',
            'anomalous high price', level='INFO')
        _test_warning(
            'data.iloc[3,2] = data.iloc[3,2] * 5',
            'anomalous high price', level='INFO')
        _test_warning(
            'data.iloc[3,2] = data.iloc[3,2] * 2',
            'anomalous high price', level='INFO')

        # extreme open price
        _test_warning(
            'data.iloc[3,0] = data.iloc[3,0] * 2;'
            + 'data.iloc[3,2] = data.iloc[3,0]',
            'anomalous open price', level='INFO')
        _test_warning(
            'data.iloc[20,0] = data.iloc[20,0] * 0.5;'
            + 'data.iloc[20,1] = data.iloc[20,0]',
            'anomalous open price', level='INFO')

        # extreme open update
        _test_warning_update(
            'new_data.iloc[-1,0] = new_data.iloc[-1,0] * 1.75;'
            + 'new_data.iloc[-1,2] = new_data.iloc[-1,0]',
            'anomalous open price', level='INFO')
        _test_warning_update(
            'new_data.iloc[-1,0] = new_data.iloc[-1,0] *  0.5;'
            + 'new_data.iloc[-1,1] = new_data.iloc[-1,0]',
            'anomalous open price', level='INFO')

    # def test_yahoo_finance_wrong_last_time(self):
    #     """Test that we correct last time if intraday."""
    #
    #     class YahooFinanceErroneous4(YahooFinance):
    #         """Modified YF that sets last time wrong."""
    #         counter = 0
    #
    #         @staticmethod
    #         def _get_data_yahoo(
    #             ticker, start='1900-01-01', end='2100-01-01'):
    #             """Modified download method."""
    #             res = YahooFinance._get_data_yahoo(
    #                 ticker, start=start, end=end)
    #             if self.counter > 0:
    #                 res.index = list(res.index)[:-1] + [
    #                     res.index[-1] - pd.Timedelta('3h')]
    #             self.counter += 1
    #             print(res)
    #             return res
    #
    #     storer = YahooFinanceErroneous4('GOOGL', base_location=self.datadir)
    #     print(storer.data)
    #     #storer.update(pd.Timedelta('0d'))
    #     #print(storer.data)


class TestMarketData(CvxportfolioTest):
    """Test MarketData methods and interface."""

    def test_market_data_downsample(self):
        """Test downsampling of market data."""
        md = DownloadedMarketData(
            ['AAPL', 'GOOG'], grace_period=self.data_grace_period,
            base_location=self.datadir)

        # TODO: better to rewrite this test
        self.strip_tz_and_hour(md)

        idx = md.returns.index

        # not doing annual because XXXX-01-01 is holiday
        freqs = ['weekly', 'monthly', 'quarterly']
        testdays = ['2023-05-01', '2023-05-01', '2022-04-01']
        periods = [['2023-05-01', '2023-05-02', '2023-05-03', '2023-05-04',
                    '2023-05-05'],
                   idx[(idx >= '2023-05-01') & (idx < '2023-06-01')],
                   idx[(idx >= '2022-04-01') & (idx < '2022-07-01')]]

        for i, freq in enumerate(freqs):

            new_md = deepcopy(md)

            # pylint: disable=protected-access
            new_md._downsample(freq)
            print(new_md.returns)
            self.assertTrue(np.isnan(new_md.returns.GOOG.iloc[0]))
            self.assertTrue(np.isnan(new_md.volumes.GOOG.iloc[0]))
            self.assertTrue(np.isnan(new_md.prices.GOOG.iloc[0]))

            if freq == 'weekly':
                print((new_md.returns.index.weekday < 2).mean())
                self.assertTrue(
                    (new_md.returns.index.weekday < 2).mean() > .95)

            if freq == 'monthly':
                print((new_md.returns.index.day < 5).mean())
                self.assertTrue((new_md.returns.index.day < 5).mean() > .95)

            self.assertTrue(
                all(md.prices.loc[testdays[i]] ==
                    new_md.prices.loc[testdays[i]]))
            self.assertTrue(np.allclose(
                md.volumes.loc[periods[i]].sum(),
                new_md.volumes.loc[testdays[i]]))
            self.assertTrue(np.allclose(
                (1 + md.returns.loc[periods[i]]).prod(),
                1 + new_md.returns.loc[testdays[i]]))

    def test_market_data_methods(self):
        """Test objects returned by serve method of MarketDataInMemory."""
        t = self.returns.index[10]
        past_returns, current_returns, past_volumes, current_volumes, \
            current_prices = self.market_data.serve(t)
        self.assertTrue(current_returns.name == t)
        self.assertTrue(current_volumes.name == t)
        self.assertTrue(current_prices.name == t)
        self.assertTrue(np.all(past_returns.index < t))
        self.assertTrue(np.all(past_volumes.index < t))

    def test_market_data_object_safety(self):
        """Test safety of internal objects of MarketDataInMemory."""
        t = self.returns.index[10]

        past_returns, current_returns, past_volumes, current_volumes, \
            current_prices = self.market_data.serve(t)

        # with warnings.catch_warnings():
        #     warnings.simplefilter("ignore")
        with self.assertRaises(ValueError):
            past_returns.iloc[-2, -2] = 2.
        with self.assertRaises(ValueError):
            current_returns.iloc[-3] = 2.
        with self.assertRaises(ValueError):
            past_volumes.iloc[-1, -1] = 2.
        with self.assertRaises(ValueError):
            current_volumes.iloc[-3] = 2.
        with self.assertRaises(ValueError):
            current_prices.iloc[-3] = 2.

        obj2 = deepcopy(self.market_data)
        # pylint: disable=protected-access
        obj2._set_read_only()

        past_returns, _, past_volumes, _, current_prices = obj2.serve(t)

        with self.assertRaises(ValueError):
            current_prices.iloc[-1] = 2.

        current_prices.loc['BABA'] = 3.

        past_returns, _, past_volumes, _, current_prices = obj2.serve(t)

        self.assertFalse('BABA' in current_prices.index)

    def test_user_provided_market_data(self):
        """Test UserProvidedMarketData."""

        used_returns = self.returns.iloc[:, :-1]
        used_returns.index = used_returns.index.tz_localize('UTC')
        used_volumes = pd.DataFrame(self.volumes, copy=True)
        t = used_returns.index[20]
        used_volumes.index = used_volumes.index.tz_localize('UTC')
        used_prices = pd.DataFrame(self.prices, copy=True)
        used_prices.index = used_prices.index.tz_localize('UTC')

        _ = UserProvidedMarketData(
            returns=used_returns, volumes=used_volumes, prices=used_prices,
            cash_key='USDOLLAR', base_location=self.datadir,
            min_history=pd.Timedelta('0d'))

        without_prices = UserProvidedMarketData(
            returns=used_returns, volumes=used_prices, cash_key='USDOLLAR',
            base_location=self.datadir, min_history=pd.Timedelta('0d'))
        _, _, past_volumes, _,  current_prices = \
            without_prices.serve(t)
        self.assertTrue(current_prices is None)

        without_volumes = UserProvidedMarketData(
            returns=used_returns, cash_key='USDOLLAR',
            base_location=self.datadir, min_history=pd.Timedelta('0d'))
        _, _, past_volumes, current_volumes, \
            current_prices = without_volumes.serve(t)

        self.assertTrue(past_volumes is None)
        self.assertTrue(current_volumes is None)

        with self.assertRaises(SyntaxError):
            UserProvidedMarketData(returns=self.returns, volumes=self.volumes,
                       prices=self.prices.iloc[:, :-1], cash_key='cash',
                       min_history=pd.Timedelta('0d'))

        with self.assertRaises(DataError):
            UserProvidedMarketData(returns=self.returns.iloc[:10],
            cash_key='cash', min_history=pd.Timedelta('50d'))

        with self.assertRaises(SyntaxError):
            UserProvidedMarketData(
                returns=self.returns,
                volumes=self.volumes.iloc[:, :-3],
                prices=self.prices, cash_key='cash',
                min_history=pd.Timedelta('0d'))

        with self.assertRaises(SyntaxError):
            used_prices = pd.DataFrame(
                self.prices, index=self.prices.index,
                columns=self.prices.columns[::-1])
            UserProvidedMarketData(returns=self.returns, volumes=self.volumes,
                       prices=used_prices, cash_key='cash',
                       min_history=pd.Timedelta('0d'))

        with self.assertRaises(SyntaxError):
            used_volumes = pd.DataFrame(
                self.volumes, index=self.volumes.index,
                columns=self.volumes.columns[::-1])
            UserProvidedMarketData(returns=self.returns, volumes=used_volumes,
                       prices=self.prices, cash_key='cash',
                       min_history=pd.Timedelta('0d'))

        with self.assertRaises(NotImplementedError):
            UserProvidedMarketData(returns=self.returns, volumes=used_volumes,
                prices=self.prices, cash_key='NOTSUPPORTED',
                min_history=pd.Timedelta('0d'))

        with self.assertRaises(ValueError):
            UserProvidedMarketData(returns=self.returns, volumes=used_volumes,
                prices=self.prices, cash_key='USDOLLAR',
                min_history=pd.Timedelta('0d'))

        md = UserProvidedMarketData(
            returns=self.returns, volumes=self.volumes,
            prices=self.prices, cash_key='cash',
            min_history=pd.Timedelta('60d'))

        # try to serve when there's not enough min_history
        with self.assertRaises(ValueError):
            md.serve(t=self.returns.index[20])

    def test_market_data_full(self):
        """Test serve method of DownloadedMarketData."""

        md = DownloadedMarketData(['AAPL', 'ZM'], base_location=self.datadir)
        assert np.all(md.full_universe == ['AAPL', 'ZM', 'USDOLLAR'])

        t = md.returns.index[-40]

        _, _, past_volumes, _, current_prices = md.serve(t)
        self.assertFalse(past_volumes is None)
        self.assertFalse(current_prices is None)

    def test_signature(self):
        """Test partial-universe signature of MarketData."""

        md = DownloadedMarketData(
            ['AAPL', 'ZM'], grace_period=self.data_grace_period,
            base_location=self.datadir)

        sig1 = md.partial_universe_signature(md.full_universe)

        md = DownloadedMarketData(
            ['AAPL', 'ZM'], grace_period=self.data_grace_period,
            trading_frequency='monthly', base_location=self.datadir)

        sig2 = md.partial_universe_signature(md.full_universe)

        self.assertFalse(sig1 == sig2)

        md = DownloadedMarketData(['AAPL', 'ZM', 'GOOG'],
            grace_period=self.data_grace_period,
            trading_frequency='monthly',
            base_location=self.datadir)

        sig3 = md.partial_universe_signature(
            pd.Index(['AAPL', 'ZM', 'USDOLLAR']))

        self.assertTrue(sig3 == sig2)

        md = DownloadedMarketData(['WM2NS'],
            datasource='Fred',
            grace_period=self.data_grace_period,
            base_location=self.datadir)

        print(md.partial_universe_signature(md.full_universe))

if __name__ == '__main__':

    unittest.main(warnings='error') # pragma: no cover



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines user-facing constraints."""

import cvxpy as cp
import numpy as np

from ..estimator import DataEstimator, Estimator
from ..forecast import (HistoricalFactorizedCovariance,
                        project_on_psd_cone_and_factorize)
from ..policies import MarketBenchmark
from .base_constraints import (Constraint, EqualityConstraint,
                               InequalityConstraint)

__all__ = [
    "LongOnly",
    "LeverageLimit",
    "LongCash",
    "DollarNeutral",
    "FactorNeutral",
    "ParticipationRateLimit",
    "MaxWeights",
    "MinWeights",
    "MaxBenchmarkDeviation",
    "MinBenchmarkDeviation",
    "NoTrade",
    "NoCash",
    "FactorMaxLimit",
    "FactorMinLimit",
    "FactorGrossLimit",
    "FixedFactorLoading",
    "MarketNeutral",
    "MinWeightsAtTimes",
    "MaxWeightsAtTimes",
    "TurnoverLimit",
    "MinCashBalance"
]

class NoCash(EqualityConstraint):
    """Require that the cash balance is zero at each period."""

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return w_plus[-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return 0


class MarketNeutral(EqualityConstraint):
    r"""Simple implementation of β- (or market-) neutrality.

    In our notation, this is

    .. math::
        {(w_t^\text{b})}^T \Sigma_t (w_t + z_t) = 0

    The benchmark portfolio weights are given by a Policy object chosen by
    the user.

    .. versionadded:: 1.2.0

        This constraint's interface has been improved: now you can pass
        any policy object as benchmark, and give parameters to the forecaster
        of :math:`\Sigma_t`.

    :param benchmark: Policy object whose target weights at each point in time
        are the benchmark weights we neutralize against. You can pass a class
        or an instance. If you pass a class it is instantiated with default
        parameters. Default is :class:`cvxportfolio.MarketBenchmark`, which are
        weights proportional to the previous year's total traded volumes.
    :type benchmark: cvx.Policy class or instance
    :param kwargs: Optional arguments passed to the initializer
        of :class:`cvxportfolio.forecast.HistoricalFactorizedCovariance`,
        like rolling window or exponential smoothing half life, for the
        estimation of the covariance matrices :math:`\Sigma_t`. Default (no
        other arguments) is to use its default parameters.
    :type kwargs: dict
    """

    def __init__(self, benchmark=MarketBenchmark, **kwargs):

        if isinstance(benchmark, type):
            benchmark = benchmark()
        self.benchmark = benchmark
        self.covariance_forecaster = HistoricalFactorizedCovariance(**kwargs)
        self._market_vector = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize parameter with size of universe.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self._market_vector = cp.Parameter(len(universe)-1)

    def values_in_time( # pylint: disable=arguments-differ
            self, past_volumes, **kwargs):
        """Update parameter with current market weights and covariance.

        :param past_volumes: Past market volumes, in units of value.
        :type past_volumes: pandas.DataFrame
        :param kwargs: Unused arguments passed to :meth:`values_in_time`.
        :type kwargs: dict
        """

        factorized_covariance = self.covariance_forecaster.current_value
        bm = self.benchmark.current_value.iloc[:-1]
        self._market_vector.value = np.array(
            factorized_covariance @ (factorized_covariance.T @ bm))

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return w_plus[:-1].T @ self._market_vector

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return 0


class TurnoverLimit(InequalityConstraint):
    r"""Turnover limit as a fraction of the portfolio value.

    The turnover is defined as half the :math:`\ell_1`-norm of
    the trade weight vector, without cash. Here we ask that it is smaller
    than some constant:

    .. math::
        \|{(z_t)}_{1:n}\|_1/2 \leq \delta.

    :param delta: We require that the turnover over each trading period
        is smaller than this value. This is either constant, expressed
        as :class:`float`, or changing in time, expressed as a
        :class:`pd.Series` with datetime index.
    :type delta: float or pd.Series
    """

    def __init__(self, delta):
        self.delta = DataEstimator(delta, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return .5 * cp.norm1(z[:-1])

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.delta.parameter


class ParticipationRateLimit(InequalityConstraint):
    """A limit on maximum trades size as a fraction of market volumes.

    :param volumes: per-stock and per-day market volume estimates, or
        constant in time
    :type volumes: pd.Series or pd.DataFrame
    :param max_fraction_of_volumes: max fraction of market volumes that
        we're allowed to trade
    :type max_fraction_of_volumes: float, pd.Series, pd.DataFrame
    """

    def __init__(self, volumes, max_fraction_of_volumes=0.05):
        self.volumes = DataEstimator(volumes, compile_parameter=True)
        self.max_participation_rate = DataEstimator(
            max_fraction_of_volumes, compile_parameter=True)
        self.portfolio_value = cp.Parameter(nonneg=True)

    def values_in_time( # pylint: disable=arguments-differ
            self, current_portfolio_value, **kwargs):
        """Update parameter with current portfolio value.

        :param current_portfolio_value: Current total value of the portfolio.
        :type current_portfolio_value: float
        :param kwargs: Unused arguments passed to :meth:`values_in_time`.
        :type kwargs: dict
        """
        self.portfolio_value.value = current_portfolio_value

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return cp.multiply(cp.abs(z[:-1]), self.portfolio_value)

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return cp.multiply(self.volumes.parameter,
                           self.max_participation_rate.parameter)


class LongOnly(InequalityConstraint):
    r"""A long only constraint.

    .. math::

        w_t + z_t \geq 0

    Imposes that at each point in time the post-trade
    weights are non-negative. By default it applies
    to all elements of the post-trade weights vector
    but you can also exclude the cash account (and let
    cash be negative).

    :param applies_to_cash: Whether the long only requirement
        also applies to the cash account.
    :type applies_to_cash: bool
    """

    def __init__(self, applies_to_cash=False):
        self.applies_to_cash = applies_to_cash

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return -(w_plus if self.applies_to_cash else w_plus[:-1])

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return 0


class NoTrade(Constraint):
    """No-trade condition on name on periods(s).

    .. note::

        This constraint object is experimental and its interface might change.

    :param asset: Symbol which can't be traded on given day(s).
    :type asset: str
    :param periods: Timestamps at which the symbol can't be traded.
    :type periods: iterable of pandas.Timestamp
    """

    def __init__(self, asset, periods):
        self.asset = asset
        self.periods = periods
        self._index = None
        self._low = None
        self._high = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize internal parameters.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self._index = (universe.get_loc if hasattr(
            universe, 'get_loc') else universe.index)(self.asset)
        self._low = cp.Parameter()
        self._high = cp.Parameter()

    def values_in_time( # pylint: disable=arguments-differ
            self, t, **kwargs):
        """Update parameters, if necessary by imposing no-trade.

        :param t: Current time.
        :type t: pandas.Timestamp
        :param kwargs: Unused arguments passed to :meth:`values_in_time`.
        :type kwargs: dict
        """
        if t in self.periods:
            self._low.value = 0.
            self._high.value = 0.
        else:
            self._low.value = -100.
            self._high.value = +100.

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile constraint to cvxpy, return list of two.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark weights.
        :type w_plus_minus_w_bm: cvxpy.Variable
        :returns: Two constraints.
        :rtype: list of cvxpy.constraints
        """
        return [z[self._index] >= self._low,
                z[self._index] <= self._high]


class LeverageLimit(InequalityConstraint):
    r"""Constraints on the leverage of the portfolio.

    In the notation of the book, this is

    .. math::

        \|{(w_t + z_t)}_{1:n}\|_1 \leq L^\text{max},

    where :math:`(w_t + z_t)` are the post-trade weights, and we
    exclude the cash account from the :math:`\ell_1` norm.

    :param limit: Constant or varying in time leverage limit
        :math:`L^\text{max}`. If varying in time it is expressed
        as a :class:`pd.Series` with datetime index.
    :type limit: float or pd.Series
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return cp.norm(w_plus[:-1], 1)

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.limit.parameter


class MinCashBalance(InequalityConstraint):
    r"""Require that the cash balance is above a threshold.

    In our notation this is

    .. math::

        {(w_t + z_t)}_{n+1} \geq c_\text{min} / v_t,

    where :math:`v_t` is the portfolio value at time :math:`t`.

    :param c_min: The miminimum cash balance required, either constant in time
        or varying, expressed in units of cash (*e.g.*, US dollars),
        not weight. Use a float if you want a constant limit, or a time-indexed
        Pandas series if you want a limit that changes in time.
    :type c_min: float or pd.Series
    """

    def __init__(self, c_min):
        self.c_min = DataEstimator(c_min)
        self.rhs = cp.Parameter()

    def values_in_time( # pylint: disable=arguments-differ
            self, current_portfolio_value, **kwargs):
        """Update parameter with current portfolio value.

        :param current_portfolio_value: Current total value of the portfolio.
        :type current_portfolio_value: float
        :param kwargs: Unused arguments passed to :meth:`values_in_time`.
        :type kwargs: dict
        """
        self.rhs.value = self.c_min.current_value/current_portfolio_value

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return -w_plus[-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return -self.rhs


class LongCash(MinCashBalance):
    r"""Require that post-trade cash account is non-negative.

    In our notation, this is

    .. math::

        {(w_t + z_t)}_{n+1} \geq 0.

    Be mindful that trading costs (if present) are deducted from the cash
    account after trading, and so the next period's cash account value
    may be negative.
    """

    def __init__(self):
        super().__init__(0.)


class DollarNeutral(EqualityConstraint):
    r"""Long-short dollar neutral strategy.

    In our notation, this is

    .. math::

        \mathbf{1}^T \max({(w_t + z_t)}_{1:n}, 0) =
            -\mathbf{1}^T \min({(w_t + z_t)}_{1:n}, 0)

    which is simply :math:`{(w_t + z_t)}_{n+1} = 1`.
    """

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return w_plus[-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return 1


class MaxWeights(InequalityConstraint):
    r"""A max limit on post-trade weights (excluding cash).

    In our notation, this is

    .. math::

        {(w_t + z_t)}_{1:n} \leq w^\text{max}

    where the limit :math:`w^\text{max}` is either a scalar or a vector, see
    below.

    :param limit: A series or number giving the weights limit. See the
        :ref:`passing-data` manual page for details on how to provide this
        data. For example, you pass a float if you want a constant limit
        for all assets at all times, a Pandas series indexed by time if you
        want a limit constant for all assets but varying in time, a Pandas
        series indexed by the assets' names if you have limits constant in time
        but different for each asset, and a Pandas dataframe indexed by time
        and with assets as columns if you have a different limit for each point
        in time and each asset. If the value changes for each asset, you should
        provide a value for each name that ever appear in a back-test; the
        data will be sliced according to the current trading universe during a
        back-test. It is fine to have missing values at certain times on assets
        that are not traded then.
    :type limit: float, pandas.Series, pandas.DataFrame
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.limit.parameter


class MinWeights(InequalityConstraint):
    r"""A min limit on post-trade weights (excluding cash).

    In our notation, this is

    .. math::

        {(w_t + z_t)}_{1:n} \geq w^\text{min}

    where the limit :math:`w^\text{min}` is either a scalar or a vector, see
    below.

    :param limit: A series or number giving the weights limit. See the
        :ref:`passing-data` manual page for details on how to provide this
        data. For example, you pass a float if you want a constant limit
        for all assets at all times, a Pandas series indexed by time if you
        want a limit constant for all assets but varying in time, a Pandas
        series indexed by the assets' names if you have limits constant in time
        but different for each asset, and a Pandas dataframe indexed by time
        and with assets as columns if you have a different limit for each point
        in time and each asset. If the value changes for each asset, you should
        provide a value for each name that ever appear in a back-test; the
        data will be sliced according to the current trading universe during a
        back-test. It is fine to have missing values at certain times on assets
        that are not traded then.
    :type limit: float, pandas.Series, pandas.DataFrame
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return -w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return -self.limit.parameter

class MaxBenchmarkDeviation(MaxWeights):
    r"""A max limit on post-trade weights minus the benchmark weights.

    In our notation, this is

    .. math::

        {(w_t + z_t - w^\text{bm}_t)}_{1:n} \leq w^\text{max}

    where the limit :math:`w^\text{max}` is either a scalar or a vector, see
    below.

    .. versionadded:: 1.1.0
        Added in version 1.1.0

    :param limit: A series or number giving the weights limit. See the
        :ref:`passing-data` manual page for details on how to provide this
        data. For example, you pass a float if you want a constant limit
        for all assets at all times, a Pandas series indexed by time if you
        want a limit constant for all assets but varying in time, a Pandas
        series indexed by the assets' names if you have limits constant in time
        but different for each asset, and a Pandas dataframe indexed by time
        and with assets as columns if you have a different limit for each point
        in time and each asset. If the value changes for each asset, you should
        provide a value for each name that ever appear in a back-test; the
        data will be sliced according to the current trading universe during a
        back-test. It is fine to have missing values at certain times on assets
        that are not traded then.
    :type limit: float, pandas.Series, pandas.DataFrame
    """

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return w_plus_minus_w_bm[:-1]


class MinBenchmarkDeviation(MinWeights):
    r"""A min limit on post-trade weights minus the benchmark weights.

    In our notation, this is

    .. math::

        {(w_t + z_t - w^\text{bm}_t)}_{1:n} \geq w^\text{min}

    where the limit :math:`w^\text{min}` is either a scalar or a vector, see
    below.

    .. versionadded:: 1.1.0
        Added in version 1.1.0

    :param limit: A series or number giving the weights limit. See the
        :ref:`passing-data` manual page for details on how to provide this
        data. For example, you pass a float if you want a constant limit
        for all assets at all times, a Pandas series indexed by time if you
        want a limit constant for all assets but varying in time, a Pandas
        series indexed by the assets' names if you have limits constant in time
        but different for each asset, and a Pandas dataframe indexed by time
        and with assets as columns if you have a different limit for each point
        in time and each asset. If the value changes for each asset, you should
        provide a value for each name that ever appear in a back-test; the
        data will be sliced according to the current trading universe during a
        back-test. It is fine to have missing values at certain times on assets
        that are not traded then.
    :type limit: float, pandas.Series, pandas.DataFrame
    """

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return -w_plus_minus_w_bm[:-1]


class MinMaxWeightsAtTimes(Estimator):
    """This class abstracts functionalities used by the two below.

    .. note::
        This class is experimental and its interface may change, or we
        may drop it.

    :param base_limit: Limit of the constraint (either ``<=`` or ``>=``).
    :type base_limit: float
    :param times: Times at which the class is active.
    :type times: iterable of pandas.Timestamp
    """

    sign = 0 # should be 1 or -1

    def __init__(self, limit, times):
        self.base_limit = limit
        self.times = times
        self.limit = None
        self.trading_calendar = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, trading_calendar, **kwargs):
        """Initialize estimator instance with updated trading_calendar.

        :param trading_calendar: Future (including current) trading calendar.
        :type trading_calendar: pandas.DatetimeIndex
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self.trading_calendar = trading_calendar
        self.limit = cp.Parameter()

    def values_in_time( # pylint: disable=arguments-differ
            self, t, mpo_step, **kwargs):
        """If target period is in sight activate constraint.

        :param t: Current time.
        :type t: pandas.Timestamp
        :param mpo_step: Which step of a
            :class:`cvxportfolio.MultiPeriodOptimization` policy we're at.
        :type mpo_step: int
        :param kwargs: Unused arguments to :meth:`values_in_time`.
        :type kwargs: dict
        """

        tidx = self.trading_calendar.get_loc(t)
        nowtidx = tidx + mpo_step
        if (nowtidx < len(self.trading_calendar)) and\
                (self.trading_calendar[nowtidx] in self.times):
            self.limit.value = self.base_limit
        else:
            self.limit.value = 100 * self.sign


class MinWeightsAtTimes(MinMaxWeightsAtTimes, InequalityConstraint):
    """Require that at certain times the weights are larger than a constant.

    .. note::
        This constraint is experimental and its interface may change, or we
        may drop it.

    :param base_limit: Minimum limit of the weights.
    :type base_limit: float
    :param times: Times at which the constraint is active.
    :type times: iterable of pandas.Timestamp
    """

    sign = -1.  # used in values_in_time of parent class

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return -w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return -self.limit


class MaxWeightsAtTimes(MinMaxWeightsAtTimes, InequalityConstraint):
    """Require that at certain times the weights are smaller than a constant.

    .. note::
        This constraint is experimental and its interface may change, or we
        may drop it.

    :param base_limit: Maximum limit of the weights.
    :type base_limit: float
    :param times: Times at which the constraint is active.
    :type times: iterable of pandas.Timestamp
    """
    sign = 1.  # used in values_in_time of parent class

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.limit


class FactorMaxLimit(InequalityConstraint):
    r"""A max limit on portfolio-wide factor (e.g. beta) exposure.

    It models the term:

    .. math::

        f_t^T {(w_t^+)}_{1:n} \leq l_t

    where :math:`f_t` is the factor exposure vector and
    :math:`l_t` the limit at time :math:`t`. It can also model a vector
    constraint

    .. math::

        F_t^T {(w_t^+)}_{1:n} \leq l_t

    where :math:`F_t` is a matrix of factor exposures and
    :math:`l_t` a vector of limits at time :math:`t`. See below for details.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time
        and has the assets names as columns, and it represents factor
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex:
        first level is time, second level are assets' names (if changing
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param limit: Factor limit, either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type limit: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(
            factor_exposure, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True,
            ignore_shape_check=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return self.factor_exposure.parameter.T @ w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.limit.parameter


class FactorMinLimit(InequalityConstraint):
    r"""A min limit on portfolio-wide factor (e.g. beta) exposure.

    It models the term:

    .. math::

        f_t^T {(w_t^+)}_{1:n} \geq l_t

    where :math:`f_t` is the factor exposure vector and
    :math:`l_t` the limit at time :math:`t`. It can also model a vector
    constraint

    .. math::

        F_t^T {(w_t^+)}_{1:n} \geq l_t

    where :math:`F_t` is a matrix of factor exposures and
    :math:`l_t` a vector of limits at time :math:`t`. See below for details.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time
        and has the assets names as columns, and it represents factor
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex:
        first level is time, second level are assets' names (if changing
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param limit: Factor limit, either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type limit: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(
            factor_exposure, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True,
            ignore_shape_check=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return -self.factor_exposure.parameter.T @ w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return -self.limit.parameter


class FactorGrossLimit(InequalityConstraint):
    r"""A gross limit on portfolio-wide factor (e.g. beta) exposure.

    It models the term:

    .. math::

        f_t^T |{(w_t^+)}_{1:n}| \leq l_t

    where :math:`f_t` is the factor exposure vector and
    :math:`l_t` the limit at time :math:`t`. It can also model a vector
    constraint

    .. math::

        F_t^T |{(w_t^+)}_{1:n}| \leq l_t

    where :math:`F_t` is a matrix of factor exposures and
    :math:`l_t` a vector of limits at time :math:`t`. See below for details.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        All elements must be non-negative.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time
        and has the assets names as columns, and it represents factor
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex:
        first level is time, second level are assets' names (if changing
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param limit: Factor limit, either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type limit: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(
            factor_exposure, non_negative=True, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True,
            ignore_shape_check=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return self.factor_exposure.parameter.T @ cp.abs(w_plus[:-1])

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.limit.parameter


class FixedFactorLoading(EqualityConstraint):
    r"""A constraint to fix portfolio loadings to a set of factors.

    This can be used to impose market neutrality,
    a certain portfolio-wide alpha, ....

    It models the term:

    .. math::

        f_t^T {(w_t^+)}_{1:n} = l_t

    where :math:`f_t` is the factor exposure vector and
    :math:`l_t` the limit at time :math:`t`. It can also model a vector
    constraint

    .. math::

        F_t^T {(w_t^+)}_{1:n} = l_t

    where :math:`F_t` is a matrix of factor exposures and
    :math:`l_t` a vector of limits at time :math:`t`. See below for details.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time
        and has the assets names as columns, and it represents factor
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex:
        first level is time, second level are assets' names (if changing
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param target: Target portfolio factor exposures,
        either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type target: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, target):
        self.factor_exposure = DataEstimator(
            factor_exposure, compile_parameter=True)
        self.target = DataEstimator(target, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile left hand side of the constraint expression."""
        return self.factor_exposure.parameter.T @ w_plus[:-1]

    def _rhs(self):
        """Compile right hand side of the constraint expression."""
        return self.target.parameter

class FactorNeutral(FixedFactorLoading):
    r"""Require neutrality with respect to certain risk factors.

    This is developed at page 35 of
    `the book <https://stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_.
    This models the term

    .. math::

        {(f_t)}^T {(w_t^+)}_{1:n} = 0

    where :math:`f_t` is the factor exposure vector. It can also model
    a vector constraint

    .. math::

        F_t^T {(w_t^+)}_{1:n} = 0

    where :math:`F_t` is a matrix of factor exposures. See below for details.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time
        and has the assets names as columns, and it represents factor
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex:
        first level is time, second level are assets' names (if changing
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure):
        super().__init__(factor_exposure=factor_exposure, target=0.)



# Copyright 2023 Enzo Busseti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains miscellaneous functions."""

import hashlib
from numbers import Number

import numpy as np
import pandas as pd

from .errors import DataError

TRUNCATE_REPR_HASH = 10  # probability of conflict is 1e-16


__all__ = ['periods_per_year_from_datetime_index', 'resample_returns',
           'flatten_heterogeneous_list', 'repr_numpy_pandas',
           'average_periods_per_year']


def set_pd_read_only(df_or_ser):
    """Set numpy array contained in dataframe or series to read only.

    This is done on data store internally before it is served to the
    policy or the simulator to ensure data consistency in case some
    element of the pipeline accidentally corrupts the data.

    This is enough to prevent direct assignement to the resulting
    dataframe. However it could still be accidentally corrupted by
    assigning to columns or indices that are not present in the
    original. We avoid that case as well by returning a wrapped
    dataframe (which doesn't copy data on creation) in
    serve_data_policy and serve_data_simulator.

    :param df_or_ser: Series or Dataframe, only numeric (better if
        homogeneous) dtype.
    :type df_or_ser: pd.Series or pd.DataFrame

    :returns: Pandas object set to read only.
    :rtype: pd.Series or pd.DataFrame
    """
    data = df_or_ser.values
    data.flags.writeable = False
    if hasattr(df_or_ser, 'columns'):
        return pd.DataFrame(data, index=df_or_ser.index,
                            columns=df_or_ser.columns)
    return pd.Series(data, index=df_or_ser.index, name=df_or_ser.name)


def average_periods_per_year(num_periods, first_time, last_time):
    """Average periods per year of a datetime index (unpacked), rounded to int.

    :param num_periods: Length of the index.
    :type num_periods: int
    :param first_time: First timestamp in the index.
    :type first_time: pandas.Timestamp
    :param last_time: Last timestamp in the index.
    :type last_time: pandas.Timestamp

    :returns: Average number of periods in a year, as implied by the index.
    :rtype: int
    """
    return int(np.round(num_periods / ((last_time - first_time) /
                                    pd.Timedelta('365.24d'))))

def periods_per_year_from_datetime_index(idx):
    """Average periods per year of a datetime index, rounded to int.

    :param idx: Datetime index, *e.g.*, of a returns dataframe.
    :type idx: pandas.DatetimeIndex

    :returns: Average number of periods in a year, as implied by the index.
    :rtype: int
    """
    return average_periods_per_year(
        num_periods=len(idx), first_time=idx[0], last_time=idx[-1])


def resample_returns(returns, periods):
    """Resample returns expressed over number of periods to single period.

    :param returns: Original returns whith are, *e.g.*, annualized, and we
        want to get as daily.
    :type returns: pandas.Series
    :param periods: Number of trading periods (*e.g.*, trading days) in the
         period of the returns (*e.g.*, 252 if annualized).
    :type periods: int

    :returns: Resampled returns.
    :rtype: pandas.Series
    """
    return np.exp(np.log(1 + returns) / periods) - 1

def make_numeric(np_or_pd):
    """Coerce Pandas or Numpy object to numeric.

    :param np_or_pd: User-provided data.
    :type np_or_pd: np.array, pd.Series, pd.DataFrame, object

    :raises DataError: If input data could not be casted to numeric.

    :returns: Same object, casted to numeric if necessary
    :rtype: np.array, pd.Series, pd.DataFrame, object
    """

    try:
        if isinstance(np_or_pd, np.ndarray):
            if not np.issubdtype(np_or_pd.dtype, np.number):
                return np_or_pd.astype(float)

        if isinstance(np_or_pd, pd.Series):
            if not np.issubdtype(np_or_pd.dtype, np.number):
                return pd.to_numeric(np_or_pd)

        if isinstance(np_or_pd, pd.DataFrame):
            if not np.all(
                [np.issubdtype(el, np.number) for el in set(np_or_pd.dtypes)]):
                return np_or_pd.astype(float)

    except ValueError as exc:
        raise DataError("Input data could not be cast to numeric.") from exc

    return np_or_pd


def flatten_heterogeneous_list(li):
    """[1, 2, 3, [4, 5]] -> [1, 2, 3, 4, 5].

    :param li: List that may contain lists, which we concatenate.
    :type li: list

    :returns: Flattened list.
    :rtype: list
    """
    return sum(([el] if not hasattr(el, '__iter__')
                else el for el in li), [])


def hash_(array_like):
    """Hash np.array.

    :param array_like: Array object of which we obtain a hash.
    :type array_like: numpy.array

    :returns: Unique hash.
    :rtype: str
    """
    return hashlib.sha256(
        bytes(str(list(array_like.flatten())), 'utf-8')).hexdigest()[
            :TRUNCATE_REPR_HASH]


def repr_numpy_pandas(nppd):
    """Unique repr of a numpy or pandas object.

    :param nppd: Numpy or pandas object of which we obtain a unique hash.
    :type nppd: numpy.array, pandas.Series, pandas.DataFrame

    :raises NotImplementedError: If the object is not supported.

    :returns: Unique hash.
    :rtype: str
    """
    if isinstance(nppd, np.ndarray):
        return f'np.array(hash={hash_(nppd)})'
    # if isinstance(nppd, np.matrix):
    #     return f'np.array(hash={hash_(nppd.A)})'
    if isinstance(nppd, pd.Series):
        return f'pd.Series(hash_values={hash_(nppd.values)}, '\
            + f'hash_index={hash_(nppd.index.to_numpy())})'
    if isinstance(nppd, pd.DataFrame):
        return f'pd.DataFrame(hash_values={hash_(nppd.values)}, '\
            + f'hash_index={hash_(nppd.index.to_numpy())}, '\
            + f'hash_columns={hash_(nppd.columns.to_numpy())})'
    raise NotImplementedError('The provided data type is not supported.')



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines :class:`BacktestResult`.

This is the object that is returned by the
:meth:`cvxportfolio.MarketSimulator.backtest`
method, and also by the same method in derived classes of
:class:`cvxportfolio.MarketSimulator`.
It contains all relevant information from a back-test and implements the logic
to compute various performance metrics, in addition to the
:meth:`BacktestResult.plot` method for producing plots
and ``__repr__`` magic method, which is invoked when the user
prints an instance.

.. versionadded:: 1.1.0
    The :attr:`BacktestResult.log` property, which returns the logs produced
    during the back-test, at level ``INFO`` or higher. It works also for
    back-tests run in parallel!
"""


from __future__ import annotations, print_function

import collections
import logging
from io import StringIO
from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .utils import periods_per_year_from_datetime_index

__all__ = ['BacktestResult']

# Module level constants, should be exposed to user (move to configuration.py?)
RECORD_LOGS = 'INFO'
LOG_FORMAT = '| %(asctime)s | %(levelname)s | process:%(process)d | %(pathname)s:%(lineno)s | %(message)s '


logger = logging.getLogger(__name__)

# def getFiscalQuarter(dt):
#     """Convert a time to a fiscal quarter."""
#     year = dt.year
#     quarter = (dt.month - 1) // 3 + 1
#     return "Q%i %s" % (quarter, year)


# pylint: disable=too-many-public-methods
class BacktestResult:
    """Store the data from a back-test and produce metrics and plots.

    :param universe: Best initial guess of the trading universe.
    :type universe: pandas.Index
    :param trading_calendar:  Best initial guess of the trading calendar.
    :type trading_calendar: pandas.DatetimeIndex
    :param costs: Simulator cost objects whose value is logged.
    :type costs: list
    """

    def __init__(self, universe, trading_calendar, costs):
        """Initialization of back-test result."""
        self._h = pd.DataFrame(index=trading_calendar,
                              columns=universe, dtype=float)
        self._u = pd.DataFrame(index=trading_calendar,
                              columns=universe, dtype=float)
        self._z = pd.DataFrame(index=trading_calendar,
                              columns=universe, dtype=float)
        self.costs = {cost.__class__.__name__: pd.Series(
            index=trading_calendar, dtype=float) for cost in costs}
        self._policy_times = pd.Series(index=trading_calendar, dtype=float)
        self._simulator_times = pd.Series(index=trading_calendar, dtype=float)
        self._cash_returns = pd.Series(index=trading_calendar, dtype=float)
        self._benchmark_returns = pd.Series(
            index=trading_calendar, dtype=float)
        self._current_universe = pd.Index(universe)
        self._indexer = np.arange(len(universe), dtype=int)

        # record logs
        self._log = ''
        self._root_logger = logging.getLogger()

        # We modify the root logger to filter at our chosen level (or lower
        # if its was lower) and pre-existing handlers (notably the
        # stderr one) to filter at the level of the original logger, or theirs
        # if higher. We put them back to their initial state at the end.
        self._orig_rootlogger_level = self._root_logger.level
        try:
            _ = logging.getLevelNamesMapping()[RECORD_LOGS]
        # for Py < 3.11
        except AttributeError: # pragma: no cover
            _ = logging._nameToLevel[RECORD_LOGS] # pragma: no cover
        self._root_logger.setLevel(min(_, self._root_logger.level))
        self._orig_loghandlers_levels = []
        for pre_existing_handler in self._root_logger.handlers:
            self._orig_loghandlers_levels.append(pre_existing_handler.level)
            pre_existing_handler.setLevel(
                max(self._orig_rootlogger_level, pre_existing_handler.level))

        # add stream handler that we use
        self._log_stream = StringIO()
        self._log_stream_handler = logging.StreamHandler(
            stream=self._log_stream)
        self._log_stream_handler.setLevel(RECORD_LOGS)
        self._root_logger.addHandler(self._log_stream_handler)

        # logs formatting
        formatter = logging.Formatter(LOG_FORMAT)
        self._log_stream_handler.setFormatter(formatter)

    @property
    def _current_full_universe(self):
        """Helper property used by ``_change_universe``.

        :returns: Current full universe (including assets that were seen
            in the past but have been dropped).
        :rtype: pandas.Index
        """
        return self._h.columns

    def _change_universe(self, new_universe):
        """Change current universe (columns of dataframes) during back-test."""

        # print('new universe')
        # print(new_universe)
        # print('old universe')
        # print(self._current_universe)

        # if necessary, expand columns of dataframes
        if not new_universe.isin(self._current_universe).all():

            # check that cash key didn't change!
            assert new_universe[-1] == self._current_universe[-1]

            # if new universe is larger we use it as ordering
            # this is the default situation with yfinance data
            # careful (thanks gh PR #114) because we need the
            # _current_full_universe, otherwise we drop assets that
            # are not traded any more
            if self._current_full_universe.isin(new_universe).all():
                joined = new_universe

            # otherwise we lose the ordering :(
            else:
                logger.info(
                    "%s joining new universe with old",
                    self.__class__.__name__)
                joined = pd.Index(
                    # need to join with full, not current!
                    sorted(set(self._current_full_universe[:-1]
                        ).union(new_universe[:-1])))
                joined = joined.append(new_universe[-1:])

            self._h = self._h.reindex(columns = joined)
            self._u = self._u.reindex(columns = joined)
            self._z = self._z.reindex(columns = joined)

        assert new_universe.isin(self._h.columns).all()
        self._current_universe = new_universe
        self._indexer = self._h.columns.get_indexer(new_universe)

    #pylint: disable=too-many-arguments
    def _log_trading(self, t: pd.Timestamp,
        h: pd.Series[float], u: pd.Series[float],
        z: pd.Series[float], costs: Dict[str, float],
        cash_return: float, benchmark_return: float or None,
        policy_time: float, simulator_time: float):
        "Log one trading period."

        if not h.index.equals(self._current_universe):
            self._change_universe(h.index)

        tidx = self._h.index.get_loc(t)

        self._h.iloc[tidx, self._indexer] = h
        self._u.iloc[tidx, self._indexer] = u
        self._z.iloc[tidx, self._indexer] = z

        for cost in costs:
            self.costs[cost].iloc[tidx] = costs[cost]

        self._simulator_times.iloc[tidx] = simulator_time
        self._policy_times.iloc[tidx] = policy_time
        self._cash_returns.iloc[tidx] = cash_return
        if benchmark_return is not None:
            self._benchmark_returns.iloc[tidx] = benchmark_return

    def _log_final(self, t, t_next, h, extra_simulator_time):
        """Log final elements and (if necessary) clean up."""

        self._h.loc[t_next] = h
        self._simulator_times.loc[t] += extra_simulator_time
        # in case of bankruptcy
        if t_next < self._h.index[-1]:
            tidx = self._h.index.get_loc(t_next)
            self._h = self._h.iloc[:tidx+1]
            self._u = self._u.iloc[:tidx]
            self._z = self._z.iloc[:tidx]
            self._simulator_times = self._simulator_times.iloc[:tidx]
            self._policy_times = self._policy_times.iloc[:tidx]
            self._cash_returns = self._cash_returns.iloc[:tidx]
            self._benchmark_returns = self._benchmark_returns.iloc[:tidx]

        # logging output
        self._log = self._log_stream.getvalue()

        # put back root logger and handlers to initial state
        self._root_logger.setLevel(self._orig_rootlogger_level)
        self._root_logger.removeHandler(self._log_stream_handler)
        for i, handler in enumerate(self._root_logger.handlers):
            handler.setLevel(self._orig_loghandlers_levels[i])

        # delete logging helpers (unnecessary?)
        del self._log_stream_handler
        self._log_stream.close()
        del self._log_stream

    #
    # General backtest information
    #

    @property
    def logs(self):
        """Logs from the policy, simulator, market data server, ....

        :return: Logs produced during the back-test, newline separated.
        :rtype: str
        """
        return self._log

    @property
    def policy_times(self):
        """The computation time of the policy object at each period.

        :returns: Policy time in seconds at each period.
        :rtype: pandas.Series
        """
        return pd.Series(self._policy_times)

    @property
    def simulator_times(self):
        """The computation time of the simulator object at each period.

        :returns: Simulator time in seconds at each period.
        :rtype: pandas.Series
        """
        return pd.Series(self._simulator_times)

    @property
    def cash_returns(self):
        """Per-period returns on cash (*i.e.*, the risk-free rate).

        :returns: All cash returns.
        :rtype: pandas.Series
        """
        return pd.Series(self._cash_returns)

    @property
    def benchmark_returns(self):
        """Benchmark returns per period (if the policy has a benchmark).

        :returns: All benchmark returns, if defined, else ``nan``.
        :rtype: pandas.Series
        """
        return pd.Series(self._benchmark_returns)

    @property
    def cash_key(self):
        """The name of the cash unit used (e.g., USDOLLAR).

        :returns: Name of the cash accounting unit.
        :rtype: str
        """
        return self._h.columns[-1]

    @property
    def periods_per_year(self):
        """Average trading periods per year in this backtest (rounded).

        :returns: Average periods per year.
        :rtype: int
        """
        return periods_per_year_from_datetime_index(self._h.index)

    #
    # Basic portfolio variables, defined in Chapter 2
    #

    @property
    def h(self):
        """The portfolio (holdings) at each trading period (including the end).

        :returns: Holdings at each period, includes cash account.
        :rtype: pandas.DataFrame
        """
        return pd.DataFrame(self._h)

    @property
    def u(self):
        """The portfolio trade vector at each trading period.

        :returns: Trades at each period.
        :rtype: pandas.DataFrame
        """
        return pd.DataFrame(self._u)

    @property
    def z(self):
        """The portfolio trade weights at each trading period.

        :returns: Trades weights at each period.
        :rtype: pandas.DataFrame
        """
        return (self.u.T / self.v).T

    @property
    def z_policy(self):
        r"""The trade weights requested by the policy at each trading period.

        This is different from the trade weights :math:`z` because the
        :class:`MarketSimulator` instance may change it by enforcing
        the self-financing condition (recalculates cash value), rounding
        trades to integer number of shares, canceling trades on assets whose
        volume is zero for the day, :math:`\ldots`.

        :returns: Trades weights requested by the policy at each period.
        :rtype: pandas.DataFrame
        """
        return pd.DataFrame(self._z)

    @property
    def v(self):
        """The total value (or NAV) of the portfolio at each period.

        :returns: Total value at each period.
        :rtype: pandas.Series
        """
        return self.h.sum(axis=1)

    @property
    def initial_value(self):
        """The initial value (or NAV) of the portfolio.

        :returns: Value at the start of the back-test.
        :rtype: float
        """
        return self.v.iloc[0]

    @property
    def final_value(self):
        """The portfolio value (or NAV) at the end of the back-test.

        :returns: Final value of the portfolio.
        :rtype: float
        """
        return self.v.iloc[-1]

    @property
    def profit(self):
        """The total profit (PnL) in this backtest.

        :returns: Total profit.
        :rtype: float
        """
        return self.v.iloc[-1] - self.v.iloc[0]

    @property
    def w(self):
        """The weights of the portfolio at each period.

        :returns: Portfolio weights at each period.
        :rtype: pandas.DataFrame
        """
        return (self.h.T / self.v).T

    @property
    def h_plus(self):
        """The post-trade portfolio (holdings) at each period.

        :returns: Post-trade holdings at each period.
        :rtype: pandas.DataFrame
        """
        return self.h.loc[self.u.index] + self.u

    @property
    def w_plus(self):
        """The post-trade weights of the portfolio at each period.

        :returns: Post-trade weights at each period.
        :rtype: pandas.DataFrame
        """
        return (self.h_plus.T / self.v).T

    @property
    def leverage(self):
        r"""Leverage of the portfolio at each period.

        This is defined as:

        .. math::

            \| {(h_t)}_{1:n} \|_1 / v_t,

        where :math:`h_t` is the portfolio (the holdings) at time :math:`t`,
        we exclude the cash account from the :math:`\ell_1` norm,
        and :math:`v_t` is the total value (NAV) of the portfolio
        at time :math:`t`.

        :returns: Leverage at each period.
        :rtype: pandas.Series
        """
        return np.abs(self.w.iloc[:, :-1]).sum(1)

    @property
    def turnover(self):
        r"""The turnover of the portfolio at each period.

        This is defined as:

        .. math::

            \| {(u_t)}_{1:n} \|_1 / (2 v_t),

        where :math:`u_t` are the portfolio trades at time :math:`t`,
        we exclude the cash account from the :math:`\ell_1` norm,
        and :math:`v_t` is the total value (NAV) of the portfolio
        at time :math:`t`.

        :returns: Turnover at each period.
        :rtype: pandas.Series
        """
        return np.abs(self.u.iloc[:, :-1]).sum(axis=1) / (
            2*self.v.loc[self.u.index])

    @property
    def returns(self):
        r"""The portfolio returns at each period.

        This is defined as:

        .. math::

            R_t^\text{p} = \frac{v_{t+1} - v_t}{v_t}

        in terms of the portfolio value (NAV).

        :returns: Portfolio returns at each period.
        :rtype: pandas.Series
        """
        val = self.v
        return pd.Series(
            data=val.values[1:] / val.values[:-1] - 1, index=val.index[:-1]
        )

    #
    # Absolute metrics, defined in Chapter 3 Section 1
    #

    @property
    def average_return(self):
        r"""The average realized return :math:`\overline{R^\text{p}}`.

        :returns: Average portfolio return.
        :rtype: float
        """
        return np.mean(self.returns)

    @property
    def annualized_average_return(self):
        r"""The average realized return, annualized.

        :returns: Average portfolio return, annualized.
        :rtype: float
        """
        return self.average_return * self.periods_per_year

    @property
    def growth_rates(self):
        r"""The growth rate (or log-return) of the portfolio at each period.

        This is defined as:

        .. math::

            G^\text{p}_t = \log (v_{t+1} / v_t) = \log(1 + R^\text{p}_t).

        :returns: Growth rate of the portfolio value at each period.
        :rtype: pandas.Series
        """
        return np.log(self.returns + 1)

    @property
    def average_growth_rate(self):
        r"""The average portfolio growth rate :math:`\overline{G^\text{p}}`.

        :returns: Average growth rate.
        :rtype: float
        """
        return np.mean(self.growth_rates)

    @property
    def annualized_average_growth_rate(self):
        r"""The average portfolio growth rate, annualized.

        :returns: Average growth rate, annualized.
        :rtype: float
        """
        return self.average_growth_rate * self.periods_per_year

    @property
    def volatility(self):
        """Realized volatility (standard deviation of the portfolio returns).

        :returns: Volatility.
        :rtype: float
        """
        return np.std(self.returns)

    @property
    def annualized_volatility(self):
        """Realized volatility, annualized.

        :returns: Volatility, annualized.
        :rtype: float
        """
        return self.volatility * np.sqrt(self.periods_per_year)

    @property
    def quadratic_risk(self):
        """Quadratic risk, square of the realized volatility.

        :returns: Quadratic risk.
        :rtype: float
        """
        return self.volatility ** 2

    @property
    def annualized_quadratic_risk(self):
        """Quadratic risk, annualized.

        :returns: Quadratic risk, annualized.
        :rtype: float
        """
        return self.quadratic_risk * self.periods_per_year

    #
    # Metrics relative to benchmark, defined in Chapter 3 Section 2
    #

    @property
    def active_returns(self):
        """Portfolio returns minus benchmark returns (if defined by policy).

        :returns: Active returns at each period if benchmark is defined,
            else ``nan``.
        :rtype: pandas.Series
        """
        return self.returns - self.benchmark_returns

    @property
    def average_active_return(self):
        r"""The average active return :math:`\overline{R^\text{a}}`.

        :returns: Average active portfolio return if benchmark is defined,
            else ``nan``.
        :rtype: float
        """
        return np.mean(self.active_returns)

    @property
    def annualized_average_active_return(self):
        """The average active return, annualized.

        :returns: Average active portfolio return, annualized. If benchmark is
            not defined, ``nan``.
        :rtype: float
        """
        return self.average_active_return * self.periods_per_year

    @property
    def active_volatility(self):
        """Active volatility (standard deviation of the active returns).

        :returns: Average active volatility if benchmark is defined,
            else ``nan``.
        :rtype: float
        """
        return np.std(self.active_returns)

    @property
    def annualized_active_volatility(self):
        """Annualized active volatility.

        :returns: Average active volatility, annualized. If benchmark is
            not defined, ``nan``.
        :rtype: float
        """
        return self.active_volatility * np.sqrt(self.periods_per_year)

    @property
    def excess_returns(self):
        """Excess portfolio returns with respect to the cash returns.

        :returns: Excess returns at each period.
        :rtype: pandas.Series
        """
        return self.returns - self.cash_returns

    @property
    def average_excess_return(self):
        r"""The average excess return :math:`\overline{R^\text{e}}`.

        :returns: Average excess portfolio return.
        :rtype: float
        """
        return np.mean(self.excess_returns)

    @property
    def annualized_average_excess_return(self):
        """The average excess return, annualized.

        :returns: Average excess portfolio return, annualized.
        :rtype: float
        """
        return self.average_excess_return * self.periods_per_year

    @property
    def excess_volatility(self):
        """Excess volatility (standard deviation of the excess returns).

        :returns: Average excess volatility.
        :rtype: float
        """
        return np.std(self.excess_returns)

    @property
    def annualized_excess_volatility(self):
        """Annualized excess volatility.

        :returns: Average excess volatility, annualized.
        :rtype: float
        """
        return self.excess_volatility * np.sqrt(self.periods_per_year)

    @property
    def sharpe_ratio(self):
        r"""Sharpe ratio (using annualized excess portfolio returns).

        This is defined as

        .. math::

            \text{SR} = \overline{R^\text{e}}/\sigma^\text{e}

        where :math:`\overline{R^\text{e}}` is the average excess portfolio
        return and :math:`\sigma^\text{e}` its standard deviation. Both are
        annualized.

        :returns: Sharpe Ratio.
        :rtype: float
        """
        return self.annualized_average_excess_return / (
            self.annualized_excess_volatility + 1E-8)

    @property
    def information_ratio(self):
        r"""Information ratio (using annualized active portfolio returns).

        This is defined as

        .. math::

            \text{IR} = \overline{R^\text{a}}/\sigma^\text{a}

        where :math:`\overline{R^\text{a}}` is the average active portfolio
        return and :math:`\sigma^\text{a}` its standard deviation. Both are
        annualized.

        :returns: Information Ratio, ``nan`` if benchmark is not defined.
        :rtype: float
        """
        return self.annualized_average_active_return / (
            self.annualized_active_volatility + 1E-8)

    @property
    def excess_growth_rates(self):
        r"""The growth rate of the portfolio, relative to cash.

        This is defined as:

        .. math::

            G^\text{e}_t = \log(1 + R^\text{e}_t)

        where :math:`R^\text{e}_t` are the excess portfolio returns.

        :returns: Excess growth rates at each period.
        :rtype: pandas.Series
        """
        return np.log(self.excess_returns + 1)

    @property
    def active_growth_rates(self):
        r"""The growth rate of the portfolio, relative to benchmark.

        This is defined as:

        .. math::

            G^\text{a}_t = \log(1 + R^\text{a}_t)

        where :math:`R^\text{a}_t` are the active portfolio returns.

        :returns: Active growth rates at each period. If benchmark is
            not defined, ``nan``.
        :rtype: pandas.Series
        """
        return np.log(self.active_returns + 1)

    @property
    def average_excess_growth_rate(self):
        r"""The average excess growth rate :math:`\overline{G^\text{e}}`.

        :returns: Average excess portfolio growth rates.
        :rtype: float
        """
        return np.mean(self.excess_growth_rates)

    @property
    def annualized_average_excess_growth_rate(self):
        """The average excess growth rate, annualized.

        :returns: Average excess portfolio growth rates, annualized.
        :rtype: float
        """
        return self.average_excess_growth_rate * self.periods_per_year

    @property
    def average_active_growth_rate(self):
        r"""The average active growth rate :math:`\overline{G^\text{a}}`.

        :returns: Average active portfolio growth rates. If benchmark is not
            defined, ``nan``.
        :rtype: float
        """
        return np.mean(self.active_growth_rates)

    @property
    def annualized_average_active_growth_rate(self):
        """The average active growth rate, annualized.

        :returns: Average active portfolio growth rates, annualized. If
            benchmark is not defined, ``nan``.
        :rtype: float
        """
        return self.average_active_growth_rate * self.periods_per_year

    @property
    def drawdown(self):
        """The drawdown of the portfolio value over time.

        :returns: Drawdown of portfolio value at each period.
        :rtype: pandas.Series
        """
        return -(1 - (self.v / self.v.cummax()))

    # TODO: decide if keeping any of these or throw

    # @staticmethod
    # def _print_growth_rate(gr):
    #     """Transform growth rate into return and pretty-print it.
    #
    #     Either prints in basis points, percentage, or multiplication.
    #     """
    #     ret = np.exp(gr)-1
    #     if np.abs(ret) < 0.005:
    #         return f'{int(ret*1E4):d}bp'
    #     if np.abs(gr) < 1:
    #         return f'{ret*100:.2f}%'
    #     return f'{1+ret:.1f}X'

    #
    # def _growth_to_return(self, growth_rates):
    #     """Convert growth to annualized percentage return."""
    #     return 100 * (np.exp(self.PPY * growth) - 1)

    # def get_quarterly_returns(self, benchmark=None):
    #     """The annualized returns for each fiscal quarter."""
    #     ret = self.growth_rates
    #     quarters = ret.groupby(getFiscalQuarter).aggregate(np.mean)
    #     return self._growth_to_return(quarters)
    #
    # def get_best_quarter(self, benchmark=None):
    #     ret = self.get_quarterly_returns(benchmark)
    #     return (ret.argmax(), ret.max())
    #
    # def get_worst_quarter(self, benchmark=None):
    #     ret = self.get_quarterly_returns(benchmark)
    #     return (ret.argmin(), ret.min())

    # @property
    # def trading_days(self):
    #     """The fraction of days with nonzero turnover."""
    #     return (self.turnover.values > 0).sum() / self.turnover.size

    #
    # Results presentation
    #

    def plot(self, show=True, how_many_weights=7):
        """Make plot and show it.

        :param show: if True, call ``matplotlib.Figure.show``, helpful when
            running in the interpreter.
        :type show: bool
        :param how_many_weights: How many assets' weights are shown in the
            weights plots. The ones with largest average absolute value
            are chosen.
        :type how_many_weights: int
        """

        # US Letter size
        fig, axes = plt.subplots(3, figsize=(8.5, 11), layout='constrained')
        fig.suptitle('Back-test result')

        # value
        self.v.plot(label='Portfolio value', ax=axes[0])
        axes[0].set_ylabel(self.cash_key)
        axes[0].set_yscale('log')
        axes[0].legend()
        axes[0].grid(True, linestyle='--', which="both")

        # weights
        biggest_weights = np.abs(self.w).mean(
            ).sort_values().iloc[-how_many_weights:].index
        self.w[biggest_weights].plot(ax=axes[1])
        axes[1].set_ylabel(f'Largest {how_many_weights} weights')
        axes[1].grid(True, linestyle='--')

        # leverage / turnover
        self.leverage.plot(ax=axes[2], linestyle='--',
                           color='k', label='Leverage')
        self.turnover.plot(ax=axes[2], linestyle='-',
                           color='r', label='Turnover')
        axes[2].legend()
        axes[2].grid(True, linestyle='--')

        if show:
            fig.show() # pragma: no cover

    def __repr__(self):
        """Print the class instance."""

        stats = collections.OrderedDict({
            "Universe size": self.h.shape[1],
            "Initial timestamp": self.h.index[0],
            "Final timestamp": self.h.index[-1],
            "Number of periods": self.u.shape[0],
            f"Initial value ({self.cash_key})": f"{self.initial_value:.3e}",
            f"Final value ({self.cash_key})": f"{self.final_value:.3e}",
            f"Profit ({self.cash_key})": f"{self.profit:.3e}",
            ' '*4: '',
            "Avg. return (annualized)":
                f"{100 * self.annualized_average_return:.1f}%",
            "Volatility (annualized)":
                f"{100 * self.annualized_volatility:.1f}%",
            "Avg. excess return (annualized)":
                f"{100 * self.annualized_average_excess_return:.1f}%",
            "Avg. active return (annualized)":
                f"{100 * self.annualized_average_active_return:.1f}%",
            "Excess volatility (annualized)":
                f"{100 * self.annualized_excess_volatility:.1f}%",
            "Active volatility (annualized)":
                f"{100 * self.annualized_active_volatility:.1f}%",
            ' '*5: '',
            "Avg. growth rate (annualized)":
                f"{100*self.annualized_average_growth_rate:.1f}%",
            "Avg. excess growth rate (annualized)":
                f"{100*self.annualized_average_excess_growth_rate:.1f}%",
            "Avg. active growth rate (annualized)":
                f"{100*self.annualized_average_active_growth_rate:.1f}%",
        })

        if len(self.costs):
            stats[' '*6] = ''
        for cost in self.costs:
            stats[f'Avg. {cost}'] = \
                f"{(self.costs[cost]/self.v).mean()*1E4:.0f}bp"
            stats[f'Max. {cost}'] = \
                f"{(self.costs[cost]/self.v).max()*1E4:.0f}bp"

        stats.update(collections.OrderedDict({
            ' '*7: '',
            "Sharpe ratio": f"{self.sharpe_ratio:.2f}",
            "Information ratio": f"{self.information_ratio:.2f}",
            ' '*8: '',
            "Avg. drawdown": f"{self.drawdown.mean() * 100:.1f}%",
            "Min. drawdown": f"{self.drawdown.min() * 100:.1f}%",
            "Avg. leverage": f"{self.leverage.mean() * 100:.1f}%",
            "Max. leverage": f"{self.leverage.max() * 100:.1f}%",
            "Avg. turnover": f"{self.turnover.mean() * 100:.1f}%",
            "Max. turnover": f"{self.turnover.max() * 100:.1f}%",
            ' '*9: '',
            "Avg. policy time": f"{self.policy_times.mean():.3f}s",
            "Avg. simulator time": f"{self.simulator_times.mean():.3f}s",
            "Total time":
                f"{self.simulator_times.sum() + self.policy_times.sum():.3f}s",
            }))

        if np.all(np.isnan(self.active_returns)):
            del stats["Avg. active return (annualized)"]
            del stats["Active volatility (annualized)"]
            del stats["Avg. active growth rate (annualized)"]
            del stats["Information ratio"]

        content = pd.Series(stats).to_string()
        lenline = len(content.split('\n')[0])

        return '\n' + '#'*lenline + '\n' + content + '\n' + '#'*lenline + '\n'



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the hyper-parameters interface."""

import unittest

import cvxportfolio as cvx
from cvxportfolio.tests import CvxportfolioTest

GAMMA_RISK_RANGE = [.5, 1., 2., 5., 10.]
GAMMA_COST_RANGE = [0., .1, .2, .5, 1., 2., 5., 10.]

class GammaRisk(cvx.RangeHyperParameter):
    """Multiplier of a risk term."""

    def __init__(self, values_range=GAMMA_RISK_RANGE, current_value=1.):
        super().__init__(values_range, current_value)


class GammaTrade(cvx.RangeHyperParameter):
    """Multiplier of a transaction cost term."""

    def __init__(self, values_range=GAMMA_COST_RANGE, current_value=1.):
        super().__init__(values_range, current_value)


class TestHyperparameters(CvxportfolioTest):
    """Test hyper-parameters interface."""

    def test_repr(self):
        """Test the repr method."""
        obj = cvx.ReturnsForecast() - cvx.Gamma() * cvx.FullCovariance()\
            - cvx.Gamma() * cvx.StocksTransactionCost()

        self.assertTrue(str(obj) ==
            'ReturnsForecast(r_hat=HistoricalMeanReturn(half_life=inf,'
            + ' rolling=inf), decay=1.0)'
            + '- Gamma(current_value=1.0) * FullCovariance('
            + 'Sigma=HistoricalFactorizedCovariance(half_life=inf,'
            + ' rolling=inf, kelly=True))'
            + '- Gamma(current_value=1.0) * StocksTransactionCost(a=0.0, '
            + 'b=1.0, market_volumes=VolumeHatOrRealized('
            + 'volume_hat=HistoricalMeanVolume(half_life=inf, '
            + "rolling=Timedelta('365 days 05:45:36'))), "
            + "sigma=HistoricalStandardDeviation(half_life=inf, "
            + "rolling=Timedelta('365 days 05:45:36'), kelly=True), "
            + 'exponent=1.5, pershare_cost=0.005)')

        print(cvx.Gamma() * cvx.Gamma())
        print(cvx.Gamma() - cvx.Gamma())
        print(cvx.Gamma() - (2 * cvx.Gamma()))
        print(cvx.Gamma() - (-2 * cvx.Gamma()))
        print(cvx.Gamma() - (cvx.Gamma() * -2))
        print(cvx.Gamma() - (cvx.Gamma() * -2))
        print(cvx.Gamma() * -1 * (cvx.Gamma() * -2))
        print(2 * cvx.Gamma() - 3 * cvx.Gamma())
        print(2 * cvx.Gamma() * ( cvx.ReturnsForecast()
            - 3 * cvx.Gamma() * cvx.FullCovariance()))
        print(cvx.Gamma() * 2 * ( cvx.ReturnsForecast()
            - cvx.Gamma() * (-3) *  cvx.FullCovariance()))

    def test_basic_hyper_parameters(self):
        """Test simple syntax and errors."""
        gamma = GammaRisk(current_value=1)

        self.assertTrue((-gamma).current_value == -1)
        self.assertTrue((gamma * .5).current_value == .5)
        self.assertTrue((.5 * gamma).current_value == .5)
        self.assertTrue((gamma).current_value == 1)

        cvx.SinglePeriodOptimization(GammaRisk() * cvx.FullCovariance())

        with self.assertRaises(SyntaxError):
            cvx.SinglePeriodOptimization(GammaRisk * cvx.FullCovariance())

        cvx.SinglePeriodOptimization(-GammaRisk() * cvx.FullCovariance())

    def test_range_hyper_parameter(self):
        """Test range hyperparameter."""

        gamma = GammaRisk(current_value=1)
        self.assertTrue((gamma).current_value == 1)
        gamma._decrement()
        self.assertTrue((gamma).current_value == .5)
        with self.assertRaises(IndexError):
            gamma._decrement()

        gamma = GammaRisk(current_value=10)
        self.assertTrue((gamma).current_value == 10)
        with self.assertRaises(IndexError):
            gamma._increment()

    def test_hyper_parameters_algebra(self):
        """Test algebra of HPs objects."""
        grisk = GammaRisk(current_value=1)
        gtrade = GammaRisk(current_value=.5)

        self.assertTrue((grisk + gtrade).current_value == 1.5)
        self.assertTrue((grisk * gtrade).current_value == .5)
        self.assertTrue((2 * grisk * gtrade).current_value == 1)
        self.assertTrue((2 * grisk * gtrade + grisk).current_value == 2)
        self.assertTrue((1.9 * grisk * gtrade + grisk).current_value == 1.95)
        self.assertTrue((grisk + 2*gtrade).current_value == 2)
        self.assertTrue((grisk/2 + 2*gtrade).current_value == 1.5)
        self.assertTrue((grisk/2 + 2 * (gtrade + gtrade/2)).current_value == 2)

    def test_collect_hyper_parameters(self):
        """Test collect hyperparameters."""

        pol = cvx.SinglePeriodOptimization(GammaRisk() * cvx.FullCovariance())

        res = pol.collect_hyperparameters()
        print(res)
        self.assertTrue(len(res) == 1)

        pol = cvx.SinglePeriodOptimization(
            - GammaRisk() * cvx.FullCovariance()
            - GammaTrade() * cvx.TransactionCost())

        res = pol.collect_hyperparameters()
        print(res)
        self.assertTrue(len(res) == 2)

        pol = cvx.SinglePeriodOptimization(
            -(GammaRisk() + .5 * GammaRisk()) * cvx.FullCovariance()
            - GammaTrade() * cvx.TransactionCost())

        res = pol.collect_hyperparameters()
        print(res)
        self.assertTrue(len(res) == 3)


if __name__ == '__main__':

    unittest.main() # pragma: no cover



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the data and parameter estimator objects."""

import unittest

import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.errors import (DataError, MissingAssetsError,
                                 MissingTimesError, NaNError)
from cvxportfolio.estimator import DataEstimator


class PlaceholderCallable:
    """Simple class with callback."""

    # pylint: disable=too-few-public-methods
    # pylint: disable=missing-return-doc
    # pylint: disable=missing-return-type-doc
    # pylint: disable=missing-param-doc

    def __init__(self, value):
        self.value = value

    def values_in_time(self, *args, **kwargs):
        """Callback used when passing this to DataEstimator."""
        return self.value


class TestEstimator(unittest.TestCase):
    """Test base Estimator objects."""

    def test_callable(self):
        """Test DataEstimator with an internal Estimator."""
        estimator = DataEstimator(PlaceholderCallable(1.0))
        time = pd.Timestamp("2022-01-01")
        self.assertEqual(estimator.values_in_time_recursive(
            t=time, current_portfolio_value=1000), 1.0)

        estimator = DataEstimator(PlaceholderCallable(np.nan))
        with self.assertRaises(NaNError):
            estimator.values_in_time_recursive(
                t=time, current_portfolio_value=1000)

        data = np.arange(10.0)
        estimator = DataEstimator(PlaceholderCallable(data))
        self.assertTrue(
            np.all(estimator.values_in_time_recursive(
                t=time, current_portfolio_value=1000) == data))

        with self.assertRaises(ValueError):
            estimator.simulate_recursive(t=time, current_portfolio_value=1000)

        data[1] = np.nan
        with self.assertRaises(NaNError):
            estimator.values_in_time_recursive(
                t=time, current_portfolio_value=1000)

    def test_scalar(self):
        """Test DataEstimator with a scalar."""
        time = pd.Timestamp("2022-01-01")

        estimator = DataEstimator(1.0)
        self.assertTrue(estimator.values_in_time_recursive(t=time) == 1.0)
        estimator = DataEstimator(1)
        self.assertTrue(estimator.values_in_time_recursive(t=time) == 1.0)

        estimator = DataEstimator(np.nan)
        with self.assertRaises(NaNError):
            estimator.values_in_time_recursive(t=time)

    def test_array(self):
        """Test DataEstimator with a numpy array."""
        time = pd.Timestamp("2022-01-01")
        data = np.arange(10.0)

        estimator = DataEstimator(data)
        self.assertTrue(
            np.all(estimator.values_in_time_recursive(t=time) == data))

        data[1] = np.nan
        estimator = DataEstimator(data)
        with self.assertRaises(NaNError):
            estimator.values_in_time_recursive(t=time)

    def test_series_dataframe_notime(self):
        """Test DataEstimator with a dataframe not time-indexed."""
        time = pd.Timestamp("2022-01-01")
        data = pd.Series(np.arange(10.0))
        estimator = DataEstimator(data)
        self.assertTrue(
            np.all(estimator.values_in_time_recursive(t=time) == data.values))

        data = pd.DataFrame(np.random.randn(3, 3))
        estimator = DataEstimator(data)
        self.assertTrue(
            np.all(estimator.values_in_time_recursive(t=time) == data.values))

    def test_series_timeindex(self):
        """Test DataEstimator with time-indexed series."""
        index = pd.date_range("2022-01-01", "2022-01-30")
        print(index)
        data = pd.Series(np.arange(len(index)), index)
        estimator = DataEstimator(data)

        print(estimator.values_in_time_recursive(t="2022-01-05"))
        self.assertTrue(estimator.values_in_time_recursive(
            t="2022-01-05") == data.loc["2022-01-05"])

        with self.assertRaises(MissingTimesError):
            estimator.values_in_time_recursive(t="2022-02-05")

        estimator = DataEstimator(data, use_last_available_time=True)
        self.assertTrue(estimator.values_in_time_recursive(
            t="2022-02-05") == data.iloc[-1])

        with self.assertRaises(MissingTimesError):
            estimator.values_in_time_recursive(t="2021-02-05")

        data["2022-01-05"] = np.nan
        estimator = DataEstimator(data)
        self.assertTrue(estimator.values_in_time_recursive(
            t="2022-01-04") == data.loc["2022-01-04"])
        with self.assertRaises(NaNError):
            estimator.values_in_time_recursive(t="2022-01-05")

        estimator = DataEstimator('hello')
        with self.assertRaises(DataError):
            estimator.values_in_time_recursive(t="2022-01-05")

    def test_dataframe_timeindex(self):
        """Test DataEstimator with time-indexed dataframe."""
        index = pd.date_range("2022-01-01", "2022-01-30")
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data)

        print(estimator.values_in_time_recursive(t="2022-01-05"))
        self.assertTrue(np.all(estimator.values_in_time_recursive(
            t="2022-01-05") == data.loc["2022-01-05"]))

        with self.assertRaises(MissingTimesError):
            estimator.values_in_time_recursive(t="2021-01-05")

        estimator = DataEstimator(data, use_last_available_time=True)
        self.assertTrue(
            np.all(estimator.values_in_time_recursive(t="2022-02-05"
                ) == data.iloc[-1]))

        data.loc["2022-01-05", 3] = np.nan
        estimator = DataEstimator(data, use_last_available_time=True)
        with self.assertRaises(MissingTimesError):
            estimator.values_in_time_recursive(t="2021-01-05")

    def test_series_notime_assetselect(self):
        """Test _universe_subselect."""
        universe = ['a', 'b', 'c']
        t = pd.Timestamp('2000-01-01')

        # data includes cash acct
        data = pd.Series(range(len(universe)), index=universe)
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=universe, trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data.values)

        # data excludes cash acct
        data = pd.Series(range(len(universe)), index=universe)
        estimator = DataEstimator(data)
        estimator.initialize_estimator_recursive(
            universe=universe, trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data.values[:2])

        # shuffled universe
        estimator = DataEstimator(data.iloc[::-1])
        estimator.initialize_estimator_recursive(
            universe=universe, trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data.values[:2])

        # wrong universe
        data = pd.Series(range(len(universe)), index=universe)
        estimator = DataEstimator(data)
        estimator.initialize_estimator_recursive(
            universe=['d', 'e', 'f'], trading_calendar=[t])
        with self.assertRaises(MissingAssetsError):
            result = estimator.values_in_time_recursive(t=t)

        # selection of universe
        data = pd.Series(range(len(universe)), index=universe)
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=['b'], trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data.values[1])

    def test_ndarray_assetselect(self):
        """Test errors if ndarray is not of right size."""
        data = np.zeros((2, 3))
        t = pd.Timestamp('2000-01-01')

        # with universe of size 2
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=['a', 'b'], trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data)

        # with universe of size 3
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=['a', 'b', 'c'], trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data)

        # error with universe of size 4
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=['a', 'b', 'c', 'd'], trading_calendar=[t])
        with self.assertRaises(MissingAssetsError):
            result = estimator.values_in_time_recursive(t=t)

        # all ok if skipping check
        estimator = DataEstimator(data, data_includes_cash=True,
            ignore_shape_check=True)
        estimator.initialize_estimator_recursive(
            universe=['a', 'b', 'c', 'd'], trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        assert np.all(result == data)

    def test_dataframe_multindex(self):
        """Test DataEstimator with a multi-indexed dataframe.

        We also check that _universe_subselect works fine.
        """
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        print(data.index)
        estimator = DataEstimator(data)
        self.assertTrue(np.all(estimator.values_in_time_recursive(
            t="2022-01-05") == data.loc["2022-01-05"]))

        # use_last_avalaible_time
        estimator = DataEstimator(data, use_last_available_time=True)
        self.assertTrue(np.all(estimator.values_in_time_recursive(
            t="2022-02-05") == data.loc["2022-01-30"]))
        self.assertTrue(np.all(estimator.values_in_time_recursive(
            t="2022-01-05") == data.loc["2022-01-05"]))
        with self.assertRaises(MissingTimesError):
            estimator.values_in_time_recursive(t="2020-01-05")

        # universe subselect
        t = "2022-01-01"
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=second_level, trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        self.assertTrue(np.all(result == data.loc[t]))

        # result has same second_level as columns
        data = pd.DataFrame(np.random.randn(len(index), len(second_level)),
                            index=index, columns=second_level)
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=second_level, trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        self.assertTrue(np.all(result == data.loc[t]))

        # universe are columns
        uni = ['a', 'b']
        data = pd.DataFrame(np.random.randn(len(index), 2),
                            index=index, columns=uni)
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=uni, trading_calendar=[t])
        result = estimator.values_in_time_recursive(t=t)
        self.assertTrue(np.all(result == data.loc[t]))

        # wrong universe
        data = pd.DataFrame(np.random.randn(len(index), 2),
                            index=index, columns=uni)
        estimator = DataEstimator(data, data_includes_cash=True)
        estimator.initialize_estimator_recursive(
            universe=uni + ['c'], trading_calendar=[t])
        with self.assertRaises(MissingAssetsError):
            result = estimator.values_in_time_recursive(t=t)

        # if timeindex is not first level it is not picked up
        index = pd.MultiIndex.from_product([second_level, timeindex])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data)
        assert np.all(estimator.values_in_time_recursive(
            t="2020-01-05") == data.values)

    def test_parameter_estimator(self):
        """Test DataEstimator's cvxpy parameter."""
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        estimator = DataEstimator(data, compile_parameter=True,
            data_includes_cash=True)
        self.assertTrue(estimator.parameter is None)
        estimator.initialize_estimator_recursive(
            universe=data.columns, trading_calendar=timeindex)
        # assert hasattr(estimator, 'parameter')
        self.assertTrue(estimator.parameter is not None)
        estimator.values_in_time_recursive(t="2022-01-05")
        self.assertTrue(
            np.all(estimator.parameter.value == data.loc["2022-01-05"]))

    def test_repr_dataestimator(self):
        """Test __repr__ magic method of DataEstimator."""
        print(DataEstimator(3))
        print(DataEstimator(np.array([1, 2, 3])))
        print(DataEstimator(pd.Series([1, 2, 3])))
        print(DataEstimator(pd.DataFrame([1, 2, 3])))

    def test_repr(self):
        """Test other assorted __repr__ methods of derived objects."""
        print(cvx.FactorModelCovariance(num_factors=10))
        print(cvx.ReturnsForecast() - .5 * cvx.FullCovariance())
        print(cvx.SinglePeriodOptimization(cvx.ReturnsForecast(),
                                           [cvx.LongOnly()]))
        print(cvx.LeverageLimit(3))


if __name__ == '__main__':

    unittest.main() # pragma: no cover
